import math
import os
import socket
import tempfile
from typing import Optional, Union
from urllib.request import urlopen
from urllib.error import HTTPError, URLError


class ConstantTemporaryFile:
    def __init__(self, constant_content, **kwargs):
        self.constant_content = constant_content
        self._file = None
        self._kwargs = dict(kwargs)
        self._kwargs["mode"] = "w+b"
        self._kwargs["delete"] = False

    def __enter__(self) -> str:
        self._file = tempfile.NamedTemporaryFile(**self._kwargs)
        self._file.write(self.constant_content)
        self._file.close()
        return self._file.name

    def __exit__(self, type, value, traceback):
        if self._file and os.path.exists(self._file.name):
            os.remove(self._file.name)
        self._file = None


def int_to_bytes(n: int) -> bytes:
    number_of_bytes = int(math.ceil(n.bit_length() / 8))
    return n.to_bytes(number_of_bytes, byteorder="big")


def decode_hex(data: Optional[str]) -> Optional[bytes]:
    if data is None:
        return None
    if data.startswith("0x"):
        data = data[2:]
    return bytes.fromhex(data)


def decode_value(v: Union[str, int]) -> int:
    if isinstance(v, int):
        return v
    elif v.startswith("0x") or (
        frozenset(["a", "b", "c", "d", "e", "f"]) & frozenset(v)
    ):
        # this is a hex string
        return int(v, 16)
    else:
        # assume it is a regular int
        return int(v)


def format_hex_address(
    addr: Optional[Union[int, str]], add_0x: bool = False
) -> Optional[str]:
    if addr is None:
        return None
    if isinstance(addr, int):
        addr = "%x" % addr
    if addr.lower().startswith("0x"):
        addr = addr[2:]
    if len(addr) < 40:
        addr = "%s%s" % ("0" * (40 - len(addr)), addr)
    elif 40 < len(addr) < 64:
        # this is likely something like a transaction hash, so round up to 32 bytes:
        addr = "%s%s" % ("0" * (64 - len(addr)), addr)
    if add_0x:
        addr = "0x%s" % addr
    return addr


def webserver_is_up(url: str) -> bool:
    try:
        return urlopen(url).getcode()
    except HTTPError:
        # This means we connected
        return True
    except URLError:
        return False


def is_port_free(port: int) -> bool:
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    return sock.connect_ex(("127.0.0.1", port)) != 0


def find_open_port(starting_port: int = 1025) -> int:
    for port in range(starting_port, 65536):
        if is_port_free(port):
            return port
    return -1


def clear_directory(path: str):
    """
    Deletes the contents of a directory, but not the directory itself.
    This is safe to use on symlinked directories.
    Symlinks will be deleted, but the files and directories they point to will not be deleted.
    If `path` itself is a symlink, the symlink will be deleted.
    """
    if os.path.islink(path):
        os.unlink(path)
        return

    for dirpath, dirnames, filenames in os.walk(path, topdown=False):
        for dirname in dirnames:
            subdir = os.path.join(dirpath, dirname)
            os.rmdir(subdir)
        for filename in filenames:
            os.remove(os.path.join(dirpath, filename))


def ynprompt(prompt: str) -> bool:
    while True:
        yn = input(prompt)
        yn = yn[0:1].lower()
        if yn == "n" or yn == "":
            return False
        elif yn == "y":
            return True



from typing import Iterable, Union

CONTROL_CODES = {
    b"0": b"\0",
    b"a": b"\x07",  # alert
    b"b": b"\x08",  # backspace
    b"f": b"\x0C",  # form feed
    b"n": b"\x0A",  # newline (line feed)
    b"r": b"\x0D",  # carriage return
    b"t": b"\x09",  # horizontal tab
    b"v": b"\x0B",  # vertical tab
    b'"': b"\x22",  # double quote
    b"&": b"",  # empty string
    b"'": b"\x27",  # single quote
    b"\\": b"\x5C",  # backslash
    b"NUL": b"\0",  # null character
    b"SOH": b"\x01",  # start of heading
    b"STX": b"\x02",  # start of text
    b"ETX": b"\x03",  # end of text
    b"EOT": b"\x04",  # end of transmission
    b"ENQ": b"\x05",  # enquiry
    b"ACK": b"\x06",  # acknowledge
    b"BEL": b"\x07",  # bell
    b"BS": b"\x08",  # backspace
    b"HT": b"\x09",  # horizontal tab
    b"LF": b"\x0A",  # line feed (newline)
    b"VT": b"\x0B",  # vertical tab
    b"FF": b"\x0C",  # form feed
    b"CR": b"\x0D",  # carriage return
    b"SO": b"\x0E",  # shift out
    b"SI": b"\x0F",  # shift in
    b"DLE": b"\x10",  # data link escape
    b"DC1": b"\x11",  # device control 1
    b"DC2": b"\x12",  # device control 2
    b"DC3": b"\x13",  # device control 3
    b"DC4": b"\x14",  # device control 4
    b"NAK": b"\x15",  # negative acknowledge
    b"SYN": b"\x16",  # synchronous idle
    b"ETB": b"\x17",  # end of transmission block
    b"CAN": b"\x18",  # cancel
    b"EM": b"\x19",  # end of medium
    b"SUB": b"\x1A",  # substitute
    b"ESC": b"\x1B",  # escape
    b"FS": b"\x1C",  # file separator
    b"GS": b"\x1D",  # group separator
    b"RS": b"\x1E",  # record separator
    b"US": b"\x1F",  # unit separator
    b"SP": b"\x20",  # space
    b"DEL": b"\x7F",  # delete
    b"^@": b"\0",
    b"^[": b"\x1B",  # escape
    b"^\\": b"\x1C",  # file separator
    b"^]": b"\x1D",  # group separator
    b"^^": b"\x1E",  # record separator
    b"^_": b"\x1F",  # unit separator
}

for i in range(26):
    CONTROL_CODES[bytes([ord("^"), ord("A") + i])] = bytes([i + 1])


def decode(text: Union[str, bytes, Iterable[int]]) -> bytes:
    escaped = None
    ret = b""
    for c in text:
        if isinstance(c, str):
            c = ord(c)
        c = bytes([c])
        if escaped is not None:

            escaped += c
            if escaped in CONTROL_CODES:
                ret += CONTROL_CODES[escaped]
                escaped = None
            elif len(escaped) >= 3:
                if len(escaped) == 3:
                    # see if it is an integer in the range [0, 255]
                    try:
                        value = int(escaped)
                        if 0 <= value <= 255:
                            ret += bytes([value])
                            escaped = None
                            continue
                    except ValueError:
                        pass
                raise ValueError(f"Unknown escape sequence: {escaped!r}")
        elif c == b"\\":
            escaped = b""
        else:
            ret += c
    return ret



#!/usr/bin/env python3

import atexit
import shlex
import shutil
import subprocess
import time

from .client import RpcHttpProxy, SelfPostingClient
from .logger import PtyLogger
from .utils import is_port_free
from .etheno import ETHENO


class Ganache(RpcHttpProxy):
    def __init__(self, cmd=None, args=None, port=8546):
        super().__init__("http://127.0.0.1:%d/" % port)
        self.port = port
        if cmd is not None:
            cmd = shlex.split(cmd)
        else:
            cmd = ["/usr/bin/env", "ganache"]
        if args is None:
            args = []
        self.args = (
            cmd + ["-d", "-p", str(port), "--chain.allowUnlimitedContractSize"] + args
        )
        self.ganache = None
        self._client = None

    def start(self):
        if self.ganache:
            return
        if shutil.which("ganache") is None:
            raise ValueError(
                "`ganache` is not installed! Install it by running `npm -g i ganache`"
            )
        if self._client:
            self.ganache = PtyLogger(self._client.logger, self.args)
            self.ganache.start()

            def ganache_errored() -> int:
                if self.ganache.is_done():
                    return self.ganache.exitstatus
                return 0

        else:
            ETHENO.logger.debug(f"Running ganache: {self.args}")
            self.ganache = subprocess.Popen(
                self.args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, bufsize=1
            )

            def ganache_errored():
                try:
                    return self.ganache.wait(0.5)
                except TimeoutError:
                    return 0

        atexit.register(Ganache.stop.__get__(self, Ganache))
        # wait until Ganache has started listening:
        while is_port_free(self.port) and not self.ganache.is_done():
            time.sleep(0.25)
        retcode = ganache_errored()
        if retcode != 0:
            raise RuntimeError(
                f"{' '.join(self.args)} exited with non-zero status {retcode}"
            )

    def post(self, data):
        if self.ganache is None:
            self.start()
        return super().post(data)

    def stop(self):
        if self.ganache is not None:
            ganache = self.ganache
            self.ganache = None
            ganache.terminate()
            ganache.wait()
            if isinstance(ganache, PtyLogger):
                ganache.close()


if __name__ == "__main__":
    ganache = Ganache()
    ganache.start()


class GanacheClient(SelfPostingClient):
    def __init__(self, ganache_instance):
        super().__init__(ganache_instance)
        ganache_instance._client = self
        self.short_name = "Ganache@%d" % ganache_instance.port

    def wait_until_running(self):
        while is_port_free(self.client.port):
            time.sleep(0.25)

    def shutdown(self):
        self.client.stop()



import time

import eth_utils
from web3.auto import w3

from .client import (
    EthenoClient,
    SelfPostingClient,
    jsonrpc,
    JSONRPCError,
    DATA,
    QUANTITY,
    transaction_receipt_succeeded,
)
from .utils import decode_hex, format_hex_address, int_to_bytes


def _decode_value(value):
    if isinstance(value, int):
        return value
    try:
        return int(value, 16)
    except Exception:
        return None


def _remap_params(client, params, mapping, method, remap_data=False):
    if isinstance(params, dict):
        for key, value in params.items():
            decoded = _decode_value(value)
            if decoded is None:
                params[key] = _remap_params(
                    client, value, mapping, "%s['%s']" % (method, key)
                )
            elif decoded in mapping:
                client.logger.debug(
                    "Converting %s parameter '%s' from %x to %x"
                    % (method, key, decoded, mapping[decoded])
                )
                params[key] = format_hex_address(mapping[decoded], True)
            elif remap_data and key == "data":
                new_value = params["data"]
                for old, new in mapping.items():
                    prev = new_value
                    new_value = new_value.replace(
                        format_hex_address(old), format_hex_address(new)
                    )
                    if prev != new_value:
                        client.logger.debug(
                            "Converting %x in %s['data'] to %x" % (old, method, new)
                        )
                if new_value != params["data"]:
                    params["data"] = new_value
    elif isinstance(params, list) or isinstance(params, tuple):
        for i, p in enumerate(params):
            decoded = _decode_value(p)
            if decoded is None:
                params[i] = _remap_params(client, p, mapping, "%s['%d']" % (method, i))
            elif decoded in mapping:
                client.logger.debug(
                    "Converting %s parameter %d from %x to %x"
                    % (method, i, decoded, mapping[decoded])
                )
                params[i] = format_hex_address(mapping[decoded], True)
    else:
        decoded = _decode_value(params)
        if decoded is not None and decoded in mapping:
            client.logger.debug(
                "Converting %s from %x to %x" % (method, decoded, mapping[decoded])
            )
            return mapping[decoded]
    return params


class ChainSynchronizer(object):
    def __init__(self, client):
        if not isinstance(client, SelfPostingClient):
            raise TypeError(
                "TODO: Implement support for address synchronization on clients other than SelfPostingClients"
            )
        self.mapping = {}
        self.filter_mapping = {}
        self._old_post = getattr(client, "post")
        self._old_create_account = getattr(client, "create_account")
        self._client = client

    def create_account(self, balance=0, address=None):
        # TODO: not sure what the data field is supposed to do here
        if self._client == self._client.etheno.master_client:
            return self._old_create_account(data)
        try:
            # First, see if the client can handle creating this address:
            return self._old_create_account(balance=balance, address=address)
        except NotImplementedError:
            pass
        new_address = self._old_create_account(balance=balance, address=None)
        if address is not None and address != new_address:
            self.mapping[address] = new_address
        return new_address

    def post(self, data, *args, **kwargs):
        if self._client == self._client.etheno.master_client:
            return self._old_post(data, *args, **kwargs)

        method = data["method"]

        if method == "eth_getTransactionReceipt":
            # first, make sure the master client's transaction succeeded; if not, we can just ignore this
            if not transaction_receipt_succeeded(self._client.etheno.rpc_client_result):
                # the master client's transaction receipt command failed, so we can skip calling this client's
                return self._client.etheno.rpc_client_result
            elif _decode_value(data["params"][0]) not in self.mapping:
                # we don't know about this transaction receipt, which probably means that the transaction failed
                # on this client. So return the receipt here, because below we will block on a result:
                return self._old_post(data, *args, **kwargs)

        uninstalling_filter = None
        if "params" in data:
            data["params"] = _remap_params(
                self._client, data["params"], self.mapping, method, remap_data=True
            )
            if (
                "filter" in method.lower() and "get" in method.lower()
            ) or method == "eth_uninstallFilter":
                # we are accessing a filter by its ID, so remap the ID
                old_id = data["params"][0]
                if old_id not in self.filter_mapping:
                    self._client.logger.warn(
                        "%s called on unknown filter ID %s; ignoring..."
                        % (method, old_id)
                    )
                else:
                    self._client.logger.info(
                        "Mapping filter ID %s to %s for %s"
                        % (old_id, self.filter_mapping[old_id], method)
                    )
                    data["params"] = [self.filter_mapping[old_id]]
                if method == "eth_uninstallFilter":
                    uninstalling_filter = old_id
        ret = self._old_post(data, *args, **kwargs)
        if uninstalling_filter is not None:
            if ret["result"]:
                # the uninstall succeeded, so we no longer need to keep the mapping:
                del self.filter_mapping[uninstalling_filter]
        elif "filter" in method.lower() and "new" in method.lower() and "result" in ret:
            # a new filter was just created, so record the mapping
            self.filter_mapping[self._client.etheno.rpc_client_result["result"]] = ret[
                "result"
            ]
        elif method == "eth_sendTransaction" or method == "eth_sendRawTransaction":
            # record the transaction hash mapping
            if ret and "result" in ret and ret["result"]:
                if (
                    self._client.etheno.rpc_client_result
                    and not isinstance(
                        self._client.etheno.rpc_client_result, JSONRPCError
                    )
                    and "result" in self._client.etheno.rpc_client_result
                    and self._client.etheno.rpc_client_result["result"]
                ):
                    old_decoded = _decode_value(
                        self._client.etheno.rpc_client_result["result"]
                    )
                    new_decoded = _decode_value(ret["result"])
                    if old_decoded is not None and new_decoded is not None:
                        self._client.logger.info(
                            "Mapping transaction hash %x to %x"
                            % (old_decoded, new_decoded)
                        )
                        self.mapping[old_decoded] = new_decoded
                    elif not (old_decoded is None and new_decoded is None):
                        self._client.logger.warn(
                            "Call to %s returned %s from the master client but %s from this client; ignoring..."
                            % (
                                method,
                                self._client.etheno.rpc_client_result["result"],
                                ret["result"],
                            )
                        )
        elif method == "eth_getTransactionReceipt":
            # by this point we know that the master client has already successfully mined the transaction and returned a receipt
            # so make sure that we block until this client has also mined the transaction and returned a receipt
            while transaction_receipt_succeeded(ret) is None:
                self._client.logger.info(
                    "Waiting to mine transaction %s..." % data["params"][0]
                )
                time.sleep(5.0)
                ret = self._old_post(data, *args, **kwargs)
            # update the mapping with the address if a new contract was created
            if "contractAddress" in ret["result"] and ret["result"]["contractAddress"]:
                master_address = _decode_value(
                    self._client.etheno.rpc_client_result["result"]["contractAddress"]
                )
                our_address = _decode_value(ret["result"]["contractAddress"])
                if master_address is not None and our_address is not None:
                    self.mapping[master_address] = our_address
                elif not (master_address is None and our_address is None):
                    self._client.logger.warn(
                        "Call to %s returned %s from the master client but %s from this client; ignoring..."
                        % (
                            method,
                            self._client.etheno.rpc_client_result["result"][
                                "contractAddress"
                            ],
                            ret["result"]["contractAddress"],
                        )
                    )

        return ret


def AddressSynchronizingClient(etheno_client):
    synchronizer = ChainSynchronizer(etheno_client)

    setattr(
        etheno_client,
        "create_account",
        ChainSynchronizer.create_account.__get__(synchronizer, ChainSynchronizer),
    )
    setattr(
        etheno_client,
        "post",
        ChainSynchronizer.post.__get__(synchronizer, ChainSynchronizer),
    )

    return etheno_client


class RawTransactionSynchronizer(ChainSynchronizer):
    def __init__(self, client, accounts):
        super().__init__(client)
        self.accounts = accounts
        self._private_keys = {}
        self._account_index = -1
        self._chain_id = client.get_net_version()

    def create_account(self, balance=0, address=None):
        self._account_index += 1
        new_address = self.accounts[self._account_index].address
        self._private_keys[new_address] = int_to_bytes(
            self.accounts[self._account_index].private_key
        )
        if address is not None and address != new_address:
            self.mapping[address] = new_address
        return new_address

    def post(self, data, *args, **kwargs):
        method = data["method"]

        if method == "eth_sendTransaction":
            # This client does not support sendTransaction because it does not have any of the requisite accounts.
            # So let's manually sign the transaction and send it to the client using eth_sendRawTransaction, instead.
            params = _remap_params(
                self._client,
                dict(data["params"][0]),
                self.mapping,
                method,
                remap_data=True,
            )
            from_str = params["from"]
            from_address = int(from_str, 16)
            if from_address in self._private_keys:
                private_key = self._private_keys[from_address]
            else:
                # see if it is in self.accounts:
                for account in self.accounts:
                    if from_address == account.address:
                        private_key = account.private_key
                        break
                else:
                    raise Exception(
                        "Error: eth_sendTransaction sent from unknown address %s:\n%s"
                        % (from_str, data)
                    )
            params["chainId"] = self._client.get_net_version()
            # Workaround for a bug in web3.eth.account:
            # the signTransaction function checks to see if the 'from' field is present, and if so it validates that it
            # corresponds to the address of the private key. However, web3.eth.account doesn't perform this check case
            # insensitively, so it can erroneously fail. Therefore, set the 'from' field using the same value that
            # this call validates against:
            params["from"] = w3.eth.account.privateKeyToAccount(private_key).address
            # web3.eth.acount.signTransaction expects the `to` field to be a checksum address:
            if "to" in params:
                params["to"] = eth_utils.address.to_checksum_address(params["to"])
            transaction_count = self._client.get_transaction_count(from_address)
            params["nonce"] = transaction_count
            signed_txn = w3.eth.account.signTransaction(params, private_key=private_key)
            return super().post(
                {
                    "id": 1,
                    "jsonrpc": "2.0",
                    "method": "eth_sendRawTransaction",
                    "params": [signed_txn.rawTransaction.hex()],
                }
            )
        else:
            return super().post(data, *args, **kwargs)


def RawTransactionClient(etheno_client, accounts):
    synchronizer = RawTransactionSynchronizer(etheno_client, accounts)

    setattr(
        etheno_client,
        "create_account",
        RawTransactionSynchronizer.create_account.__get__(
            synchronizer, RawTransactionSynchronizer
        ),
    )
    setattr(
        etheno_client,
        "post",
        RawTransactionSynchronizer.post.__get__(
            synchronizer, RawTransactionSynchronizer
        ),
    )

    return etheno_client



#!/usr/bin/env python3

import threading
from threading import Condition


def is_main_thread():
    return isinstance(threading.current_thread(), threading._MainThread)


class MainThreadController(object):
    def __init__(self):
        if not is_main_thread():
            raise Exception("A controller can only be created from the main thread")
        self._main_wake_up = Condition()
        self._main_wake_up.acquire()
        self._obj = None
        self._caller_wake_up = None
        self._args = None
        self._kwargs = None
        self._return = None
        self._quit = False

    def invoke(self, obj, *args, **kwargs):
        if is_main_thread():
            return obj.__call__(*args, **kwargs)
        released = False
        self._main_wake_up.acquire()
        try:
            self._caller_wake_up = Condition()
            with self._caller_wake_up:
                self._obj = obj
                self._args = args
                self._kwargs = kwargs
                # tell the main thread to wake up
                self._main_wake_up.notify_all()
                self._main_wake_up.release()
                released = True
                self._caller_wake_up.wait()
                # tell the main thread that we received the result:
                self._caller_wake_up.notify_all()
                ret = self._return
                return ret
        finally:
            self._obj = None
            self._args = None
            self._kwargs = None
            self._caller_wake_up = None
            self._return = None
            if not released:
                self._main_wake_up.release()

    def quit(self):
        self._main_wake_up.acquire()
        try:
            self._quit = True
            self._main_wake_up.notify_all()
        finally:
            self._main_wake_up.release()

    def run(self):
        if not is_main_thread():
            raise Exception("run can only be called from the main thread!")
        from . import signals

        def signal_handler(signal, frame):
            self._quit = True

        signals.add_sigint_handler(signal_handler)
        while True:
            try:
                self._main_wake_up.wait(1.0)
            except KeyboardInterrupt:
                self._quit = True
            if self._quit:
                return
            elif self._caller_wake_up is None:
                # we timed out
                continue
            with self._caller_wake_up:
                self._return = self._obj.__call__(*self._args, **self._kwargs)
                self._caller_wake_up.notify_all()
                # wait for the calling thread to confirm it received the result
                while not self._caller_wake_up.wait(1.0):
                    if self._quit:
                        return


class MainThreadWrapper(object):
    def __init__(self, mainobj, controller):
        self._main = mainobj
        self._controller = controller

    def __call__(self, *args, **kwargs):
        ret = self._controller.invoke(self._main, *args, **kwargs)
        if id(self._main) == id(ret):
            return MainThreadWrapper(ret, self._controller)
        else:
            return ret

    def __getattribute__(self, name):
        if name == "_main" or name == "_controller":
            return object.__getattribute__(self, name)
        elif isinstance(getattr(type(self._main), name), property):
            return getattr(self._main, name)
        else:
            return MainThreadWrapper(getattr(self._main, name), self._controller)


if __name__ == "__main__":

    class MainThreadOnlyClass(object):
        def do_stuff(self):
            if not is_main_thread():
                raise Exception("ERROR!")
            return 1337

    main_thread_only = MainThreadOnlyClass()

    controller = MainThreadController()

    def dostuff(mtoc):
        print(mtoc.do_stuff())

    from threading import Thread

    thread = Thread(
        target=dostuff, args=(MainThreadWrapper(main_thread_only, controller),)
    )
    thread.start()
    controller.run()
    thread.join()



import signal


def add_handler(signal_type, handler):
    current_handler = signal.getsignal(signal_type)
    if current_handler == signal.SIG_IGN or current_handler == signal.SIG_DFL:
        current_handler = None

    def new_handler(sig_type, frame):
        if current_handler:
            current_handler(sig_type, frame)
        handler(sig_type, frame)

    signal.signal(signal_type, new_handler)


def add_sigint_handler(handler):
    add_handler(signal.SIGINT, handler)



# Code adapted from https://github.com/ethereum/eth-keyfile/blob/master/eth_keyfile/keyfile.py
# to be compatible with Parity
# Specifically, Parity requires a 256bit salt
#
# This file is licensed under the MIT License (MIT)
#
# Copyright (c) 2017 Piper Merriam
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import hashlib
import hmac
import json
import uuid

from Crypto import Random
from Crypto.Cipher import AES
from Crypto.Protocol.KDF import scrypt
from Crypto.Util import Counter

from eth_keys import keys

from eth_utils import (
    big_endian_to_int,
    decode_hex,
    encode_hex,
    int_to_big_endian,
    is_dict,
    is_string,
    keccak,
    remove_0x_prefix,
    to_dict,
)


def encode_hex_no_prefix(value):
    return remove_0x_prefix(encode_hex(value))


def load_keyfile(path_or_file_obj):
    if is_string(path_or_file_obj):
        with open(path_or_file_obj) as keyfile_file:
            return json.load(keyfile_file)
    else:
        return json.load(path_or_file_obj)


def create_keyfile_json(
    private_key, password, version=3, kdf="pbkdf2", iterations=None
):
    if version == 3:
        return _create_v3_keyfile_json(private_key, password, kdf, iterations)
    else:
        raise NotImplementedError("Not yet implemented")


def decode_keyfile_json(raw_keyfile_json, password):
    keyfile_json = normalize_keys(raw_keyfile_json)
    version = keyfile_json["version"]

    if version == 3:
        return _decode_keyfile_json_v3(keyfile_json, password)
    else:
        raise NotImplementedError("Not yet implemented")


def extract_key_from_keyfile(path_or_file_obj, password):
    keyfile_json = load_keyfile(path_or_file_obj)
    private_key = decode_keyfile_json(keyfile_json, password)
    return private_key


@to_dict
def normalize_keys(keyfile_json):
    for key, value in keyfile_json.items():
        if is_string(key):
            norm_key = key.lower()
        else:
            norm_key = key

        if is_dict(value):
            norm_value = normalize_keys(value)
        else:
            norm_value = value

        yield norm_key, norm_value


#
# Version 3 creators
#
DKLEN = 32
SCRYPT_R = 1
SCRYPT_P = 8


def _create_v3_keyfile_json(private_key, password, kdf, work_factor=None):
    salt = Random.get_random_bytes(32)

    if work_factor is None:
        work_factor = get_default_work_factor_for_kdf(kdf)

    if kdf == "pbkdf2":
        derived_key = _pbkdf2_hash(
            password,
            hash_name="sha256",
            salt=salt,
            iterations=work_factor,
            dklen=DKLEN,
        )
        kdfparams = {
            "c": work_factor,
            "dklen": DKLEN,
            "prf": "hmac-sha256",
            "salt": encode_hex_no_prefix(salt),
        }
    elif kdf == "scrypt":
        derived_key = _scrypt_hash(
            password,
            salt=salt,
            buflen=DKLEN,
            r=SCRYPT_R,
            p=SCRYPT_P,
            n=work_factor,
        )
        kdfparams = {
            "dklen": DKLEN,
            "n": work_factor,
            "r": SCRYPT_R,
            "p": SCRYPT_P,
            "salt": encode_hex_no_prefix(salt),
        }
    else:
        raise NotImplementedError("KDF not implemented: {0}".format(kdf))

    iv = big_endian_to_int(Random.get_random_bytes(16))
    encrypt_key = derived_key[:16]
    ciphertext = encrypt_aes_ctr(private_key, encrypt_key, iv)
    mac = keccak(derived_key[16:32] + ciphertext)

    address = keys.PrivateKey(private_key).public_key.to_address()

    return {
        "address": remove_0x_prefix(address),
        "crypto": {
            "cipher": "aes-128-ctr",
            "cipherparams": {
                "iv": encode_hex_no_prefix(int_to_big_endian(iv)),
            },
            "ciphertext": encode_hex_no_prefix(ciphertext),
            "kdf": kdf,
            "kdfparams": kdfparams,
            "mac": encode_hex_no_prefix(mac),
        },
        "id": str(uuid.uuid4()),
        "version": 3,
    }


#
# Verson 3 decoder
#
def _decode_keyfile_json_v3(keyfile_json, password):
    crypto = keyfile_json["crypto"]
    kdf = crypto["kdf"]

    # Derive the encryption key from the password using the key derivation
    # function.
    if kdf == "pbkdf2":
        derived_key = _derive_pbkdf_key(crypto, password)
    elif kdf == "scrypt":
        derived_key = _derive_scrypt_key(crypto, password)
    else:
        raise TypeError("Unsupported key derivation function: {0}".format(kdf))

    # Validate that the derived key matchs the provided MAC
    ciphertext = decode_hex(crypto["ciphertext"])
    mac = keccak(derived_key[16:32] + ciphertext)

    expected_mac = decode_hex(crypto["mac"])

    if not hmac.compare_digest(mac, expected_mac):
        raise ValueError("MAC mismatch")

    # Decrypt the ciphertext using the derived encryption key to get the
    # private key.
    encrypt_key = derived_key[:16]
    cipherparams = crypto["cipherparams"]
    iv = big_endian_to_int(decode_hex(cipherparams["iv"]))

    private_key = decrypt_aes_ctr(ciphertext, encrypt_key, iv)

    return private_key


#
# Key derivation
#
def _derive_pbkdf_key(crypto, password):
    kdf_params = crypto["kdfparams"]
    salt = decode_hex(kdf_params["salt"])
    dklen = kdf_params["dklen"]
    should_be_hmac, _, hash_name = kdf_params["prf"].partition("-")
    assert should_be_hmac == "hmac"
    iterations = kdf_params["c"]

    derive_pbkdf_key = _pbkdf2_hash(password, hash_name, salt, iterations, dklen)

    return derive_pbkdf_key


def _derive_scrypt_key(crypto, password):
    kdf_params = crypto["kdfparams"]
    salt = decode_hex(kdf_params["salt"])
    p = kdf_params["p"]
    r = kdf_params["r"]
    n = kdf_params["n"]
    buflen = kdf_params["dklen"]

    derived_scrypt_key = _scrypt_hash(
        password,
        salt=salt,
        n=n,
        r=r,
        p=p,
        buflen=buflen,
    )
    return derived_scrypt_key


def _scrypt_hash(password, salt, n, r, p, buflen):
    derived_key = scrypt(
        password,
        salt=salt,
        key_len=buflen,
        N=n,
        r=r,
        p=p,
        num_keys=1,
    )
    return derived_key


def _pbkdf2_hash(password, hash_name, salt, iterations, dklen):
    derived_key = hashlib.pbkdf2_hmac(
        hash_name=hash_name,
        password=password,
        salt=salt,
        iterations=iterations,
        dklen=dklen,
    )

    return derived_key


#
# Encryption and Decryption
#
def decrypt_aes_ctr(ciphertext, key, iv):
    ctr = Counter.new(128, initial_value=iv, allow_wraparound=True)
    encryptor = AES.new(key, AES.MODE_CTR, counter=ctr)
    return encryptor.decrypt(ciphertext)


def encrypt_aes_ctr(value, key, iv):
    ctr = Counter.new(128, initial_value=iv, allow_wraparound=True)
    encryptor = AES.new(key, AES.MODE_CTR, counter=ctr)
    ciphertext = encryptor.encrypt(value)
    return ciphertext


#
# Utility
#
def get_default_work_factor_for_kdf(kdf):
    if kdf == "pbkdf2":
        return 1000000
    elif kdf == "scrypt":
        return 262144
    else:
        raise ValueError("Unsupported key derivation function: {0}".format(kdf))


if __name__ == "__main__":
    from utils import int_to_bytes

    print(
        create_keyfile_json(
            int_to_bytes(
                0xDA105EBBEDE69170F1D13C70B0B82715A7CB46740779457D7EE4FBFA54BA95C1
            ),
            b"etheno",
        )
    )



import http
import inspect
import json
import time
from typing import Any, Dict, List, Optional, Set, Union
from urllib.request import Request, urlopen

from . import logger
from .utils import decode_hex, format_hex_address, webserver_is_up


def jsonrpc(**types):
    def decorator(function):
        signature = inspect.getfullargspec(function).args

        def wrapper(self, *args, **kwargs):
            rpc_kwargs = dict(kwargs)
            return_type = None
            converted_args = []
            for i, arg in enumerate(args):
                if signature[i + 1] in types:
                    converted_args.append(types[signature[i + 1]](arg))
                else:
                    converted_args.append(arg)
                rpc_kwargs[signature[i + 1]] = arg
            args = tuple(converted_args)
            kwargs = dict(kwargs)
            for arg_name, conversion in types.items():
                if arg_name == "RETURN":
                    return_type = conversion
                elif arg_name in kwargs:
                    kwargs[arg_name] = conversion(kwargs[arg_name])
            return function(self, *args, **kwargs)

        return wrapper

    return decorator


class RpcHttpProxy:
    def __init__(self, urlstring):
        self.urlstring = urlstring
        self.rpc_id = 0

    def post(self, data) -> Dict[str, Union[int, str, Dict[str, Any]]]:
        data = dict(data)
        self.rpc_id += 1
        return_id = None
        if "jsonrpc" not in data:
            data["jsonrpc"] = "2.0"
        if "id" in data:
            return_id = data["id"]
            data["id"] = self.rpc_id
        request = Request(
            self.urlstring,
            data=bytearray(json.dumps(data), "utf8"),
            headers={"Content-type": "application/json"},
        )
        ret = json.loads(urlopen(request).read())
        if return_id is not None and "id" in ret:
            ret["id"] = return_id
        return ret

    def __str__(self):
        return f"{self.__class__.__name__}<{self.urlstring}>"

    def __repr__(self):
        return f"{self.__class__.__name__}({self.urlstring})"


class JSONRPCError(RuntimeError):
    def __init__(self, client, data, result):
        super().__init__(
            "JSON RPC Error in Client %s when processing transaction:\n%s\n%s"
            % (client, data, result["error"])
        )
        self.client = client
        self.json = data
        self.result = result


def transaction_receipt_succeeded(data):
    if not (data and "result" in data and data["result"]):
        return None
    elif "contractAddress" in data["result"] and data["result"]["contractAddress"]:
        return True
    elif "blockHash" in data["result"] and data["result"]["blockHash"]:
        return True
    elif "status" not in data["result"]:
        return None
    status = data["result"]["status"]
    if status is None:
        return None
    elif not isinstance(status, int):
        status = int(status, 16)
    return status > 0


class EthenoClient:
    _etheno = None
    logger = None
    _short_name = None

    @property
    def etheno(self):
        return self._etheno

    @etheno.setter
    def etheno(self, instance):
        if self._etheno is not None:
            if instance is None:
                self._etheno = None
                return
            elif instance == self._etheno:
                return
            raise ValueError(
                "An Etheno client can only ever be associated with a single Etheno instance"
            )
        self._etheno = instance
        self.logger = logger.EthenoLogger(self.short_name, parent=self._etheno.logger)
        self.etheno_set()

    @property
    def log_directory(self) -> Optional[str]:
        """Returns a log directory that this client can use to save additional files, or None if one is not available"""
        if self.logger is None:
            return None
        else:
            return self.logger.directory

    def etheno_set(self):
        """A callback for once the etheno instance and logger for this client is set"""
        pass

    def create_account(self, balance=0, address=None):
        """A request for the client to create a new account.

        Subclasses implementing this function should raise a NotImplementedError if an address
        was provided that the client is unable to create an account at that address

        :param balance: The initial balance for the account
        :param address: The address for the account, or None if the address should be auto-generated
        :return: returns the address of the account created
        """
        raise NotImplementedError("Clients must extend this function")

    def shutdown(self):
        pass

    def wait_for_transaction(self, tx_hash):
        return None

    @property
    def short_name(self):
        if self._short_name is None:
            return str(self)
        else:
            return self._short_name

    @short_name.setter
    def short_name(self, name):
        self._short_name = name


class SelfPostingClient(EthenoClient):
    def __init__(self, client):
        self.client = client
        self._accounts: Optional[List[int]] = None
        self._created_account_index = -1
        # maintain a set of failed transactions so we know not to block on eth_getTransactionReceipt
        self._failed_transactions: Set[str] = set()

    def create_account(self, balance: int = 0, address: Optional[int] = None):
        if address is not None:
            raise NotImplementedError()
        if self._accounts is None:
            self._accounts = list(
                int(a, 16)
                for a in self.post(
                    {"id": 1, "jsonrpc": "2.0", "method": "eth_accounts"}
                )["result"]
            )
        self._created_account_index += 1
        return self._accounts[self._created_account_index]

    def wait_until_running(self):
        pass

    # TODO: need to ensure that JSON RPC calls match latest API spec
    def post(self, data: Dict[str, Any]) -> Optional[Dict[str, Any]]:
        ret = self.client.post(data)
        if ret is not None and "error" in ret:
            if "method" in data and (
                data["method"] == "eth_sendTransaction"
                or data["method"] == "eth_sendRawTransaction"
            ):
                if (
                    self.etheno.master_client != self
                    and self.etheno.rpc_client_result
                    and not isinstance(self.etheno.rpc_client_result, JSONRPCError)
                    and "result" in self.etheno.rpc_client_result
                ):
                    self.logger.error(
                        f"{self!s}: Failed transaction associated with master client transaction "
                        f"{self.etheno.rpc_client_result['result']}"
                    )
                    self._failed_transactions.add(
                        self.etheno.rpc_client_result["result"].lower()
                    )
            # TODO: Figure out a better way to handle JSON RPC errors
            raise JSONRPCError(self, data, ret)
        return ret

    def estimate_gas(self, transaction: Dict[str, Any]) -> int:
        """Estimates the gas cost for the given transaction or call

        :param transaction: a dict containing the entire transaction as if it were to be sent to `post()`
        :return: the gas cost in wei as an int
        """
        return int(
            self.post(
                {
                    "id": 1,
                    "jsonrpc": "2.0",
                    "method": "eth_estimateGas",
                    "params": transaction["params"],
                }
            )["result"],
            16,
        )

    def get_gas_price(self) -> int:
        return int(
            self.post({"id": 1, "jsonrpc": "2.0", "method": "eth_gasPrice"})["result"],
            16,
        )

    def get_net_version(self) -> int:
        return int(
            self.post({"id": 1, "jsonrpc": "2.0", "method": "net_version"})["result"],
            16,
        )

    def get_transaction_count(self, from_address) -> int:
        return int(
            self.post(
                {
                    "id": 1,
                    "jsonrpc": "2.0",
                    "method": "eth_getTransactionCount",
                    "params": [format_hex_address(from_address, True), "latest"],
                }
            )["result"],
            16,
        )

    def wait_for_transaction(self, tx_hash):
        """Blocks until the given transaction has been mined
        :param tx_hash: the transaction hash for the transaction to monitor
        :return: The transaction receipt
        """
        while True:
            request_object = self.etheno.get_transaction_receipt_request(tx_hash)
            receipt = self.post(request_object)
            if (
                tx_hash in self._failed_transactions
                or transaction_receipt_succeeded(receipt) is not None
            ):
                return receipt
            self.logger.info("Waiting to mine transaction %s..." % tx_hash)
            time.sleep(5.0)

    def __str__(self):
        return f"{self.__class__.__name__}[{self.client!s}]"

    def __repr__(self):
        return f"{self.__class__.__name__}[{self.client!r}]"


class RpcProxyClient(SelfPostingClient):
    def __init__(self, rpcurl):
        super().__init__(RpcHttpProxy(rpcurl))

    def post(self, data):
        while True:
            try:
                return super().post(data)
            except http.client.RemoteDisconnected as e:
                self.logger.warning(str(e))
                time.sleep(1.0)
                self.logger.info(f"Retrying JSON RPC call to {self.client.urlstring}")

    def is_running(self) -> bool:
        return webserver_is_up(self.client.urlstring)

    def wait_until_running(self):
        slept = 0.0
        while not self.is_running():
            time.sleep(0.25)
            slept += 0.25
            if slept % 5 == 0:
                self.logger.info("Waiting for the client to start...")


def QUANTITY(to_convert: Optional[str]) -> Optional[int]:
    if to_convert is None:
        return None
    elif to_convert[:2] == "0x":
        return int(to_convert[2:], 16)
    else:
        return int(to_convert)


def DATA(to_convert: Optional[str]) -> Optional[bytes]:
    return decode_hex(to_convert)



from collections.abc import Sequence
import copy
import json
import os

from .client import RpcProxyClient
from .genesis import make_accounts
from .logger import PtyLogger
from .utils import format_hex_address, is_port_free


class JSONRPCClient(RpcProxyClient):
    def __init__(self, name, genesis, port=8546):
        super().__init__("http://localhost:%d/" % port)
        self._basename = name
        self.short_name = "%s@%d" % (name, port)
        self.port = port
        self.genesis = copy.deepcopy(genesis)
        self.miner_account = make_accounts(1)[0]
        self.genesis["alloc"][format_hex_address(self.miner_account.address)] = {
            "balance": "0",
            "privateKey": format_hex_address(self.miner_account.private_key),
        }
        self._accounts = []
        self._created_address_index = -1
        self._runscript = []
        # These are set in self.etheno_set():
        self.genesis_file = None
        self.passwords = None
        self.datadir = None
        # This is set when self.start() is called:
        self.instance = None

    def write_genesis(self, outfile):
        outfile.write(json.dumps(self.genesis).encode("utf-8"))

    def write_passwords(self, outfile):
        for i in range(len(self.genesis["alloc"])):
            outfile.write(b"etheno\n")

    def etheno_set(self):
        super().etheno_set()
        self.datadir = os.path.join(self.log_directory, "chain_data")
        os.makedirs(self.datadir)
        with self.logger.make_logged_file(
            prefix="genesis", suffix=".json"
        ) as genesis_output:
            self.genesis_file = genesis_output.name
            self.write_genesis(genesis_output)
            genesis_output.close()
        with self.logger.make_logged_file(
            prefix=self._basename, suffix=".passwd"
        ) as password_output:
            self.passwords = password_output.name
            self.write_passwords(password_output)

    def add_to_run_script(self, command):
        if isinstance(command, Sequence):
            command = " ".join(command)
        self._runscript.append(command)

    def import_account(self, private_key):
        raise NotImplementedError()

    @property
    def accounts(self):
        for addr, bal in self.genesis["alloc"].items():
            yield int(addr, 16)

    def create_account(self, balance: int = 0, address=None):
        accounts = list(self.accounts)
        if address is None:
            self._created_address_index += 1
            if self._created_address_index >= len(accounts):
                raise Exception(
                    "Ran out of %s genesis accounts and could not create a new one!"
                    % self.short_name
                )
            return accounts[self._created_address_index]
        elif address not in accounts:
            valid_accounts = "\n".join(map(hex, accounts))
            raise ValueError(
                f"Account {address!s} did not exist in the genesis for client {self.short_name}! "
                f"Valid accounts:\n{valid_accounts}"
            )
        else:
            return address

    def get_start_command(self, unlock_accounts=True):
        raise NotImplementedError()

    def save_run_script(self):
        run_script = os.path.join(
            self.log_directory, "run_%s.sh" % self._basename.lower()
        )
        with open(run_script, "w") as f:
            script = "\n".join(self._runscript)
            f.write(script)
        # make the script executable:
        os.chmod(run_script, 0o755)

    def start(self, unlock_accounts=True):
        start_args = self.get_start_command(unlock_accounts)
        self.instance = PtyLogger(self.logger, start_args, cwd=self.log_directory)
        if self.log_directory:
            self.add_to_run_script(start_args)
            self.save_run_script()
        self.initialized()
        self.instance.start()
        self.wait_until_running()

    def initialized(self):
        """Called once the client is completely intialized but before it is started"""
        pass

    def is_running(self):
        return not is_port_free(self.port)

    def stop(self):
        if self.instance is not None:
            instance = self.instance
            self.instance = None
            instance.terminate()
            instance.wait()
            instance.close()

    def shutdown(self):
        self.stop()



import atexit
import os
import subprocess
import time

from . import logger
from .client import JSONRPCError
from .jsonrpcclient import JSONRPCClient
from .utils import format_hex_address


def ltrim_ansi(text):
    if text.startswith(logger.ANSI_RESET):
        return ltrim_ansi(text[len(logger.ANSI_RESET) :])
    elif text.startswith(logger.ANSI_BOLD):
        return ltrim_ansi(text[len(logger.ANSI_BOLD) :])
    for color in logger.CGAColors:
        ansi = logger.ANSI_COLOR % (30 + color.value)
        if text.startswith(ansi):
            return ltrim_ansi(text[len(ansi) :])
    for color in logger.CGAColors:
        ansi = f"\033[{30 + color.value}m"
        if text.startswith(ansi):
            return ltrim_ansi(text[len(ansi) :])
    return text


class GethClient(JSONRPCClient):
    def __init__(self, genesis, port=8546):
        super().__init__("Geth", genesis, port)
        atexit.register(GethClient.shutdown.__get__(self, GethClient))

    def initialized(self):
        def log(logger, message):
            msg = ltrim_ansi(message)
            if msg.startswith("ERROR"):
                logger.error(msg[5:].lstrip())
            elif msg.startswith("WARNING"):
                logger.warning(msg[7:].lstrip())
            elif msg.startswith("WARN"):
                logger.warning(msg[4:].lstrip())
            elif msg.startswith("DEBUG"):
                logger.debug(msg[5:].lstrip())
            elif msg.startswith("INFO"):
                logger.info(msg[4:].lstrip())
            else:
                logger.info(message)

        self.instance.log = log

    def etheno_set(self):
        super().etheno_set()
        try:
            args = [
                "/usr/bin/env",
                "geth",
                "init",
                self.logger.to_log_path(self.genesis_file),
                "--datadir",
                self.logger.to_log_path(self.datadir),
            ]
            self.add_to_run_script(args)
            subprocess.check_call(
                args,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
                cwd=self.log_directory,
            )
        except Exception as e:
            self.cleanup()
            raise e

    def import_account(self, private_key):
        content = format_hex_address(private_key).encode("utf-8") + bytes([ord("\n")])
        import_dir = os.path.join(self.log_directory, "private_keys")
        keyfile = self.logger.make_constant_logged_file(
            content, prefix="private", suffix=".key", dir=import_dir
        )
        while True:
            args = [
                "/usr/bin/env",
                "geth",
                "account",
                "import",
                "--datadir",
                self.logger.to_log_path(self.datadir),
                "--password",
                self.logger.to_log_path(self.passwords),
                self.logger.to_log_path(keyfile),
            ]
            self.add_to_run_script(args)
            geth = subprocess.Popen(
                args,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                cwd=self.log_directory,
            )
            if geth.wait() == 0:
                return
            # This sometimes happens with geth, I have no idea why, so just try again

    def post(self, data):
        # geth takes a while to unlock all of the accounts, so check to see if that caused an error and just wait a bit
        while True:
            try:
                return super().post(data)
            except JSONRPCError as e:
                if (
                    e.result["error"]["code"] == -32000
                    and "authentication needed" in e.result["error"]["message"]
                ):
                    self.logger.info(
                        "Waiting for Geth to finish unlocking our accounts..."
                    )
                    time.sleep(3.0)
                else:
                    raise e

    def get_start_command(self, unlock_accounts=True):
        if self.logger.log_level == logger.CRITICAL:
            verbosity = 0
        elif self.logger.log_level == logger.ERROR:
            verbosity = 1
        elif self.logger.log_level == logger.WARNING:
            verbosity = 2
        elif self.logger.log_level == logger.DEBUG:
            verbosity = 3
        else:
            verbosity = 4
        base_args = [
            "/usr/bin/env",
            "geth",
            "--nodiscover",
            "--rpc",
            "--rpcport",
            "%d" % self.port,
            "--networkid",
            "%d" % self.genesis["config"]["chainId"],
            "--datadir",
            self.logger.to_log_path(self.datadir),
            "--mine",
            "--etherbase",
            format_hex_address(self.miner_account.address),
            f"--verbosity={verbosity}",
            "--minerthreads=1",
        ]
        if unlock_accounts:
            addresses = filter(
                lambda a: a != format_hex_address(self.miner_account.address),
                map(format_hex_address, self.genesis["alloc"]),
            )
            unlock_args = [
                "--unlock",
                ",".join(addresses),
                "--password",
                self.passwords,
            ]
        else:
            unlock_args = []
        return base_args + unlock_args



import atexit
import json
import os
import tempfile

from .client import JSONRPCError
from .genesis import geth_to_parity
from .jsonrpcclient import JSONRPCClient
from .keyfile import create_keyfile_json
from .utils import find_open_port, format_hex_address, int_to_bytes


def make_config(genesis_path, base_path, port, accounts, password_file, **kwargs):
    return """[parity]
public_node = false
no_download = false
no_consensus = false
no_persistent_txqueue = false

chain = "{genesis_path}"
base_path = "{base_path}"
db_path = "{base_path}/chains"
keys_path = "{base_path}/keys"

[account]
unlock = [{account_addresses}]
password = ["{password_file}"]

[network]
port = {port}
min_peers = 1
max_peers = 1
id = {chainId}
discovery = false

[rpc]
disable = false
port = {rpc_port}
interface = "local"
apis = ["web3", "eth", "pubsub", "net", "parity", "parity_pubsub", "traces", "rpc", "shh", "shh_pubsub"]
hosts = ["none"]

[websockets]
disable = true

[ipc]
disable = true

[secretstore]
disable = true

[ipfs]
enable = false

[mining]
author = "{miner}"
engine_signer = "{miner}"
force_sealing = true
reseal_on_txs = "all"
#reseal_min_period = 4000
#reseal_max_period = 60000
#gas_floor_target = "4700000"
#gas_cap = "6283184"
#tx_queue_gas = "off"
#tx_gas_limit = "6283184"
#tx_time_limit = 500 #ms
#remove_solved = false
#notify_work = ["http://localhost:3001"]
refuse_service_transactions = false

[footprint]
tracing = "auto"
pruning = "auto"
pruning_history = 64
pruning_memory = 32
cache_size_db = 128
cache_size_blocks = 8
cache_size_queue = 40
cache_size_state = 25
cache_size = 128 # Overrides above caches with total size
fast_and_loose = false
db_compaction = "ssd"
fat_db = "auto"
scale_verifiers = true
num_verifiers = 6

[snapshots]
disable_periodic = false

[misc]
logging = "own_tx=trace"
log_file = "{log_path}"
color = true
""".format(
        genesis_path=genesis_path,
        base_path=base_path,
        port=find_open_port(30303),
        rpc_port=port,
        log_path=kwargs.get("log_path", "%s/parity.log" % base_path),
        chainId=kwargs.get("chainId", 1),
        miner=format_hex_address(accounts[-1], True),
        account_addresses=", ".join(
            map(lambda s: '"0x%s"' % s, map(format_hex_address, accounts))
        ),
        password_file=password_file,
    ).encode(
        "utf-8"
    )


class ParityClient(JSONRPCClient):
    def __init__(self, genesis, port=8546):
        super().__init__("Parity", genesis, port)
        self._unlock_accounts = True

        self.config = None

        atexit.register(ParityClient.shutdown.__get__(self, ParityClient))

    def etheno_set(self):
        super().etheno_set()
        self.import_account(self.miner_account.private_key)
        self.config = self.logger.make_constant_logged_file(
            make_config(
                genesis_path=self.logger.to_log_path(self.genesis_file),
                base_path=self.logger.to_log_path(self.datadir),
                port=self.port,
                chainId=self.genesis["config"]["chainId"],
                accounts=tuple(self.accounts),
                password_file=self.logger.to_log_path(self.passwords),
            ),
            prefix="config",
            suffix=".toml",
        )

    def write_passwords(self, outfile):
        outfile.write(b"etheno")

    def write_genesis(self, outfile):
        parity_genesis = geth_to_parity(self.genesis)
        parity_genesis["genesis"]["author"] = format_hex_address(
            self.miner_account.address, True
        )
        outfile.write(json.dumps(parity_genesis).encode("utf-8"))

    def import_account(self, private_key):
        keyfile = create_keyfile_json(int_to_bytes(private_key), b"etheno")
        keyfile_json = json.dumps(keyfile)
        keysdir = os.path.join(self.datadir, "keys", "etheno")
        os.makedirs(keysdir, exist_ok=True)
        output = tempfile.NamedTemporaryFile(
            prefix="account", suffix=".key", dir=keysdir, delete=False
        )
        try:
            output.write(keyfile_json.encode("utf-8"))
        finally:
            output.close()
        if self.log_directory is None:
            self._tempfiles.append(output)

    def unlock_account(self, account):
        addr = format_hex_address(account, True)
        self.logger.info("Unlocking Parity account %s..." % addr)
        return self.post(
            {
                "id": addr,
                "jsonrpc": "2.0",
                "method": "personal_unlockAccount",
                "params": [addr, "etheno", None],  # Unlock the account for one day
            }
        )

    def post(self, data, unlock_if_necessary=None):
        if unlock_if_necessary is None:
            unlock_if_necessary = self._unlock_accounts
        try:
            return super().post(data)
        except JSONRPCError as e:
            if (
                unlock_if_necessary
                and "data" in e.result["error"]
                and e.result["error"]["data"].lower() == "notunlocked"
            ):
                self.unlock_account(int(data["params"][0]["from"], 16))
                return self.post(data, unlock_if_necessary=False)
            else:
                raise e

    def get_start_command(self, unlock_accounts=True):
        return [
            "/usr/bin/env",
            "parity",
            "--config",
            self.logger.to_log_path(self.config),
            "--fast-unlock",
            "--jsonrpc-apis=all",
        ]

    def start(self, unlock_accounts=True):
        self._unlock_accounts = unlock_accounts
        super().start(unlock_accounts=unlock_accounts)



from enum import Enum
import os

from .client import JSONRPCError, SelfPostingClient
from .etheno import EthenoPlugin


class DifferentialTest(object):
    def __init__(self, tester, test_name, success, message=""):
        self.tester = tester
        self.test_name = test_name
        self.message = message
        self.success = success
        self.tester.logger.make_constant_logged_file(
            self.message,
            prefix=["FAILED", "PASSED"][self.success.value],
            suffix=".log",
            dir=os.path.join(self.tester.logger.directory, self.test_name),
        )

    def __str__(self):
        return "[%s] %s\t%s" % (self.test_name, self.success, self.message)

    __repr__ = __str__


class TestResult(Enum):
    FAILED = 0
    PASSED = 1


class DifferentialTester(EthenoPlugin):
    def __init__(self):
        self._transactions_by_hash = {}
        self._unprocessed_transactions = set()
        self.tests = {}
        self._printed_summary = False

    def add_test_result(self, result):
        if result.test_name not in self.tests:
            self.tests[result.test_name] = {}
        if result.success not in self.tests[result.test_name]:
            self.tests[result.test_name][result.success] = []
        self.tests[result.test_name][result.success].append(result)

    def after_post(self, data, client_results):
        method = data["method"]

        # First, see if any of the clients returned an error. If one did, they all should!
        clients_with_errors = tuple(
            i
            for i, result in enumerate(client_results)
            if isinstance(result, JSONRPCError)
        )
        clients_without_errors = tuple(
            sorted(
                frozenset(range(len(client_results))) - frozenset(clients_with_errors)
            )
        )

        if clients_with_errors:
            clients = [self.etheno.master_client] + self.etheno.clients
            if clients_without_errors:
                test = DifferentialTest(
                    self,
                    "JSON_RPC_ERRORS",
                    TestResult.FAILED,
                    "%s executed JSON RPC call %s with no errors, but %s executed the same transaction with errors:\n%s"
                    % (
                        ", ".join(
                            str(clients[client]) for client in clients_without_errors
                        ),
                        data,
                        ", ".join(
                            str(clients[client]) for client in clients_with_errors
                        ),
                        "\n".join(
                            str(client_results[client])
                            for client in clients_with_errors
                        ),
                    ),
                )
            else:
                test = DifferentialTest(
                    self,
                    "JSON_RPC_ERRORS",
                    TestResult.PASSED,
                    "All clients executed JSON RPC call %s with errors" % data,
                )
            self.add_test_result(test)
            self.logger.error(test.message)
            return
        else:
            self.add_test_result(
                DifferentialTest(
                    self,
                    "JSON_RPC_ERRORS",
                    TestResult.PASSED,
                    "All clients executed transaction %s without error" % data,
                )
            )

        master_result = client_results[0]
        if method == "eth_sendTransaction" or method == "eth_sendRawTransaction":
            if (
                not isinstance(master_result, JSONRPCError)
                and "result" in master_result
                and master_result["result"]
            ):
                self._unprocessed_transactions.add(master_result["result"])
                self._transactions_by_hash[master_result["result"]] = data
        elif method == "eth_getTransactionReceipt":
            if master_result and "result" in master_result and master_result["result"]:
                # mark that we have processed the receipt for this transaction:
                if data["params"][0] in self._unprocessed_transactions:
                    self._unprocessed_transactions.remove(data["params"][0])

                if (
                    "contractAddress" in master_result["result"]
                    and master_result["result"]["contractAddress"]
                ):
                    # the master client created a new contract
                    # so make sure that all of the other clients did, too
                    for client, client_data in zip(
                        self.etheno.clients, client_results[1:]
                    ):
                        created = False
                        try:
                            created = client_data["result"]["contractAddress"]
                        except Exception:
                            pass
                        if not created:
                            test = DifferentialTest(
                                self,
                                "CONTRACT_CREATION",
                                TestResult.FAILED,
                                f"{self.etheno.master_client} created a contract for transaction {data['params'][0]}, but {client} did not",
                            )
                            self.add_test_result(test)
                            self.logger.error(test.message)
                        else:
                            self.add_test_result(
                                DifferentialTest(
                                    self,
                                    "CONTRACT_CREATION",
                                    TestResult.PASSED,
                                    "client %s transaction %s"
                                    % (client, data["params"][0]),
                                )
                            )
                if (
                    "gasUsed" in master_result["result"]
                    and master_result["result"]["gasUsed"]
                ):
                    # make sure each client used the same amount of gas
                    master_gas = int(master_result["result"]["gasUsed"], 16)
                    for client, client_data in zip(
                        self.etheno.clients, client_results[1:]
                    ):
                        gas_used = 0
                        try:
                            gas_used = int(client_data["result"]["gasUsed"], 16)
                        except Exception:
                            pass
                        if gas_used != master_gas:
                            test = DifferentialTest(
                                self,
                                "GAS_USAGE",
                                TestResult.FAILED,
                                f"""Transaction {data['params'][0]} used:

    {hex(master_gas)} gas in {self.etheno.master_client} but
    {hex(gas_used)} gas in {client}

while mining this transaction:

{self._transactions_by_hash.get(data['params'][0], 'UNKNOWN TRANSACTION')}""",
                            )
                            self.add_test_result(test)
                            self.logger.error(test.message)
                        else:
                            self.add_test_result(
                                DifferentialTest(
                                    self,
                                    "GAS_USAGE",
                                    TestResult.PASSED,
                                    "client %s transaction %s used 0x%x gas"
                                    % (client, data["params"][0], gas_used),
                                )
                            )

                # we have processed this transaction, so no need to keep its original arguments around:
                if data["params"][0] in self._transactions_by_hash:
                    del self._transactions_by_hash[data["params"][0]]

    def finalize(self):
        unprocessed = self._unprocessed_transactions
        self._unprocessed_transactions = set()
        for tx_hash in unprocessed:
            self.logger.info(
                "Requesting transaction receipt for %s to check differentials..."
                % tx_hash
            )
            if not isinstance(self.etheno.master_client, SelfPostingClient):
                self.logger.warn(
                    "The DifferentialTester currently only supports master clients that extend from SelfPostingClient, but %s does not; skipping checking transaction(s) %s"
                    % (self.etheno.master_client, ", ".join(unprocessed))
                )
                return
            while True:
                receipt = self.etheno.post(
                    {
                        "jsonrpc": "2.0",
                        "method": "eth_getTransactionReceipt",
                        "params": [tx_hash],
                    }
                )
                # if this post is successful, it will trigger the `after_post` callback above
                # where were check for the differentials
                if "result" in receipt and receipt["result"]:
                    break
                # The transaction is still pending
                time.sleep(3.0)

    def shutdown(self):
        # super().shutdown() should automatically call self.finalize()
        super().shutdown()
        if self.tests and not self._printed_summary:
            self._printed_summary = True
            ret = "\nDifferential Test Summary:\n\n"
            for test in sorted(self.tests):
                ret += "    %s\n" % test
                total = sum(map(len, self.tests[test].values()))
                for result in self.tests[test]:
                    ret += "        %s\t%d / %d\n" % (
                        result,
                        len(self.tests[test][result]),
                        total,
                    )
                ret += "\n"
            self.logger.info(ret)



from setuptools import setup, find_packages

setup(
    name="etheno",
    description="Etheno is a JSON RPC multiplexer, differential fuzzer, and test framework integration tool.",
    url="https://github.com/trailofbits/etheno",
    author="Trail of Bits",
    version="0.3.1",
    packages=find_packages(),
    python_requires=">=3.7",
    install_requires=[
        "ptyprocess",
        "pysha3>=1.0.2",
        # TODO: identify what is the oldest flask version that the new shutdown mechanism is compatible with
        "flask",
        # Pinning web3 to a low version to prevent conflicts with other packages
        "web3>=3.16.4",
        # Contextual version conflicts between eth-hash, eth-utils, eth-rlp, and rusty-rlp
        # This works only if `--platform linux/amd64` is set since rusty-rlp==0.1.15 is not available for ARM architectures
        # This is super hacky but it works for now
        # This is likely going to cause conflicts with other packages :(
        "eth-hash>=0.3.1,<0.4.0",
        "eth-utils==1.10.0",
        "eth-rlp<0.3.0",
        "setuptools",
    ],
    # rusty-rlp==0.1.15 has to be downloaded as a tarball
    dependency_links=[
        "https://github.com/cburgdorf/rusty-rlp/archive/refs/tags/0.1.15.tar.gz"
    ],
    entry_points={"console_scripts": ["etheno = etheno.__main__:main"]},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Topic :: Security",
        "Topic :: Software Development :: Testing",
    ],
)



from web3.auto import w3

from .utils import format_hex_address


class Account(object):
    def __init__(self, address, balance=None, private_key=None):
        self._address = address
        self.balance = balance
        self._private_key = private_key

    @property
    def address(self):
        return self._address

    @property
    def private_key(self):
        return self._private_key


def make_genesis(
    network_id=0x657468656E6F,
    difficulty=20,
    gas_limit=200000000000,
    accounts=None,
    byzantium_block=0,
    dao_fork_block=0,
    homestead_block=0,
    eip150_block=0,
    eip155_block=0,
    eip158_block=0,
    constantinople_block=None,
):
    if accounts:
        alloc = {
            format_hex_address(acct.address): {
                "balance": "%d" % acct.balance,
                "privateKey": format_hex_address(acct.private_key),
            }
            for acct in accounts
        }
    else:
        alloc = {}

    ret = {
        "config": {
            "chainId": network_id,
            "byzantiumBlock": byzantium_block,
            "daoForkBlock": dao_fork_block,
            "homesteadBlock": homestead_block,
            "eip150Block": eip150_block,
            "eip155Block": eip155_block,
            "eip158Block": eip158_block,
        },
        "difficulty": "%d" % difficulty,
        "gasLimit": "%d" % gas_limit,
        "alloc": alloc,
    }

    if constantinople_block is not None:
        ret["config"]["constantinopleBlock"] = constantinople_block

    return ret


def geth_to_parity(genesis):
    """Converts a Geth style genesis to Parity style"""
    ret = {
        "name": "etheno",
        "engine": {
            "instantSeal": None,
            # 'Ethash': {
            #     'params': {
            #         'minimumDifficulty': "0x%s" % genesis['difficulty'],
            #         'difficultyBoundDivisor': '0x100000000',
            #         'homesteadTransition': 0,
            #         'eip150Transition': 0,
            #         'eip160Transition': 0,
            #         'eip161abcTransition': 0,
            #         'eip161dTransition': 0,
            #     }
            # }
        },
        "genesis": {
            "seal": {
                "generic": "0x0"
                # 'ethereum': {
                #    'nonce': '0x0000000000000042',
                #    'mixHash': '0x0000000000000000000000000000000000000000000000000000000000000000'
                # }
            },
            "difficulty": "0x%s" % genesis["difficulty"],
            "gasLimit": "0x%s" % genesis["gasLimit"],
            "author": list(genesis["alloc"])[-1],
        },
        "params": {
            "networkID": "0x%x" % genesis["config"]["chainId"],
            "maximumExtraDataSize": "0x20",
            "minGasLimit": "0x%s" % genesis["gasLimit"],
            "gasLimitBoundDivisor": "1",
            "eip150Transition": "0x%x" % genesis["config"]["eip150Block"],
            "eip160Transition": "0x0",
            "eip161abcTransition": "0x0",
            "eip161dTransition": "0x0",
            "eip155Transition": "0x%x" % genesis["config"]["eip155Block"],
            "eip98Transition": "0x7fffffffffffff",
            # 'eip86Transition': '0x7fffffffffffff',
            "maxCodeSize": 24576,
            "maxCodeSizeTransition": "0x0",
            "eip140Transition": "0x0",
            "eip211Transition": "0x0",
            "eip214Transition": "0x0",
            "eip658Transition": "0x0",
            "wasmActivationTransition": "0x0",
        },
        "accounts": dict(genesis["alloc"]),
    }

    if "constantinopleBlock" in genesis["config"]:
        block = "0x%x" % genesis["config"]["constantinopleBlock"]
        ret["params"]["eip145Transition"] = block
        ret["params"]["eip1014Transition"] = block
        ret["params"]["eip1052Transition"] = block

    return ret


DEFAULT_PRIVATE_KEYS = [
    0xF2F48EE19680706196E2E339E5DA3491186E0C4C5030670656B0E0164837257D,
    0x5D862464FE9303452126C8BC94274B8C5F9874CBD219789B3EB2128075A76F72,
    0xDF02719C4DF8B9B8AC7F551FCB5D9EF48FA27EEF7A66453879F4D8FDC6E78FB1,
    0xFF12E391B79415E941A94DE3BF3A9AEE577AED0731E297D5CFA0B8A1E02FA1D0,
    0x752DD9CF65E68CFABA7D60225CBDBC1F4729DD5E5507DEF72815ED0D8ABC6249,
    0xEFB595A0178EB79A8DF953F87C5148402A224CDF725E88C0146727C6ACEADCCD,
    0x83C6D2CC5DDCF9711A6D59B417DC20EB48AFD58D45290099E5987E3D768F328F,
    0xBB2D3F7C9583780A7D3904A2F55D792707C345F21DE1BACB2D389934D82796B2,
    0xB2FD4D29C1390B71B8795AE81196BFD60293ADF99F9D32A0AFF06288FCDAC55F,
    0x23CB7121166B9A2F93AE0B7C05BDE02EAE50D64449B2CBB42BC84E9D38D6CC89,
]


def make_accounts(num_accounts, default_balance=None):
    ret = []
    if num_accounts > len(DEFAULT_PRIVATE_KEYS):
        raise Exception("TODO: Too many accounts")
    for i in range(num_accounts):
        acct = w3.eth.account.from_key(DEFAULT_PRIVATE_KEYS[i])
        ret.append(
            Account(
                address=int(acct.address, 16),
                private_key=int(acct.privateKey.hex(), 16),
                balance=default_balance,
            )
        )
    return ret



import enum
import logging
import os
import tempfile
import threading
import time
from typing import Callable, List, Optional, Union, Any

import ptyprocess

CRITICAL = logging.CRITICAL
ERROR = logging.ERROR
WARNING = logging.WARNING
INFO = logging.INFO
DEBUG = logging.DEBUG
NOTSET = logging.NOTSET


class CGAColors(enum.Enum):
    BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(8)


ANSI_RESET = "\033[0m"
ANSI_COLOR = "\033[1;%dm"
ANSI_BOLD = "\033[1m"

LEVEL_COLORS = {
    CRITICAL: CGAColors.MAGENTA,
    ERROR: CGAColors.RED,
    WARNING: CGAColors.YELLOW,
    INFO: CGAColors.GREEN,
    DEBUG: CGAColors.CYAN,
    NOTSET: CGAColors.BLUE,
}

# TODO: seems like this function can be removed, no references?
def formatter_message(message: str, use_color: bool = True) -> str:
    if use_color:
        message = message.replace("$RESET", RESET_SEQ).replace("$BOLD", BOLD_SEQ)
    else:
        message = message.replace("$RESET", "").replace("$BOLD", "")
    return message


class ComposableFormatter:
    def __init__(self, *args, **kwargs):
        if len(args) == 1 and not isinstance(args[0], str):
            self._parent_formatter = args[0]
        else:
            self._parent_formatter = self.new_formatter(*args, **kwargs)

    def new_formatter(self, *args, **kwargs):
        return logging.Formatter(*args, **kwargs)

    def __getattr__(self, name):
        return getattr(self._parent_formatter, name)


class ColorFormatter(ComposableFormatter):
    def reformat(self, fmt: str) -> str:
        for color in CGAColors:
            fmt = fmt.replace("$%s" % color.name, ANSI_COLOR % (30 + color.value))
        fmt = fmt.replace("$RESET", ANSI_RESET)
        fmt = fmt.replace("$BOLD", ANSI_BOLD)
        return fmt

    @staticmethod
    def remove_color(fmt: str) -> str:
        for color in CGAColors:
            fmt = fmt.replace("$%s" % color.name, "")
        fmt = fmt.replace("$RESET", "")
        fmt = fmt.replace("$BOLD", "")
        fmt = fmt.replace("$LEVELCOLOR", "")
        return fmt

    def new_formatter(self, fmt: str, *args, **kwargs) -> logging.Formatter:
        if "datefmt" in kwargs:
            kwargs["datefmt"] = self.reformat(kwargs["datefmt"])
        return super().new_formatter(self.reformat(fmt), *args, **kwargs)

    def format(self, *args, **kwargs) -> str:
        levelcolor = LEVEL_COLORS.get(args[0].levelno, LEVEL_COLORS[NOTSET])
        ret = self._parent_formatter.format(*args, **kwargs)
        ret = ret.replace("$LEVELCOLOR", ANSI_COLOR % (30 + levelcolor.value))
        ret = ret.replace("\n", self.reformat("$RESET $BOLD$BLUE\\$RESET\n"), 1)
        ret = ret.replace("\n", self.reformat("\n$RESET$BOLD$BLUE> $RESET"))
        return ret


class NonInfoFormatter(ComposableFormatter):
    _vanilla_formatter = logging.Formatter()

    def format(self, *args, **kwargs) -> str:
        if args and args[0].levelno == INFO:
            return self._vanilla_formatter.format(*args, **kwargs)
        else:
            return self._parent_formatter.format(*args, **kwargs)


ETHENO_LOGGERS = {}

_LOGGING_GETLOGGER = logging.getLogger


def getLogger(name: str):
    if name in ETHENO_LOGGERS:
        # TODO: Only enable this if Etheno was run as a standalone application
        ret = ETHENO_LOGGERS[name]
    else:
        ret = _LOGGING_GETLOGGER(name)
    return ret


logging.getLogger = getLogger


class EthenoLogger:
    DEFAULT_FORMAT = (
        "$RESET$LEVELCOLOR$BOLD%(levelname)-8s $BLUE[$RESET$WHITE%(asctime)14s$BLUE$BOLD]$NAME$RESET "
        "%(message)s"
    )

    def __init__(
        self,
        name: str,
        log_level: Optional[int] = None,
        parent: Optional["EthenoLogger"] = None,
        cleanup_empty: bool = False,
        displayname: Optional[str] = None,
    ):
        if name in ETHENO_LOGGERS:
            raise Exception(
                f"An EthenoLogger instance for name {name} already exists: {ETHENO_LOGGERS[name]}"
            )
        ETHENO_LOGGERS[name] = self
        self._directory: Optional[str] = None
        self.parent: Optional[EthenoLogger] = parent
        self.cleanup_empty: bool = cleanup_empty
        self.children: List[EthenoLogger] = []
        self._descendant_handlers = []
        if displayname is None:
            self.displayname: str = name
        else:
            self.displayname = displayname
        if log_level is None:
            if parent is None:
                raise ValueError(
                    "A logger must be provided a parent if `log_level` is None"
                )
            log_level = parent.log_level
        self._log_level: int = log_level
        self._logger: logging.Logger = _LOGGING_GETLOGGER(name)
        self._handlers: List[logging.Handler] = [logging.StreamHandler()]
        if log_level is not None:
            self.log_level = log_level
        formatter = ColorFormatter(
            self.DEFAULT_FORMAT.replace("$NAME", self._name_format()),
            datefmt="%m$BLUE-$WHITE%d$BLUE|$WHITE%H$BLUE:$WHITE%M$BLUE:$WHITE%S",
        )
        if self.parent is None:
            formatter = NonInfoFormatter(formatter)
        else:
            parent._add_child(self)
        self._handlers[0].setFormatter(formatter)  # type: ignore
        self._logger.addHandler(self._handlers[0])
        self._tmpdir = None

    def close(self):
        for child in self.children:
            child.close()
        if self.cleanup_empty:
            # first, check any files that handlers have created:
            for h in self._handlers:
                if isinstance(h, logging.FileHandler):
                    if h.stream is not None:
                        log_path = h.stream.name
                        if os.path.exists(log_path) and os.stat(log_path).st_size == 0:
                            h.close()
                            os.remove(log_path)
            # next, check if the output directory can be cleaned up
            if self.directory:
                for dirpath, dirnames, filenames in os.walk(
                    self.directory, topdown=False
                ):
                    if (
                        len(dirnames) == 0
                        and len(filenames) == 0
                        and dirpath != self.directory
                    ):
                        os.rmdir(dirpath)
        if self._tmpdir is not None:
            self._tmpdir.cleanup()
            self._tmpdir = None

    @property
    def directory(self) -> Optional[str]:
        return self._directory

    def _add_child(self, child):
        if child in self.children or any(
            c for c in self.children if c.name == child.name
        ):
            raise ValueError(
                "Cannot double-add child logger %s to logger %s"
                % (child.name, self.name)
            )
        self.children.append(child)
        if self.directory is not None:
            child.save_to_directory(
                os.path.join(self.directory, child.name.replace(os.sep, "-"))
            )
        else:
            child._tmpdir = tempfile.TemporaryDirectory()
            child.save_to_directory(child._tmpdir.name)
        parent = self
        while parent is not None:
            for handler in self._descendant_handlers:
                child.addHandler(handler, include_descendants=True)
            parent = parent.parent

    def _name_format(self):
        if self.parent is not None and self.parent.parent is not None:
            ret = self.parent._name_format()
        else:
            ret = ""
        return ret + "[$RESET$WHITE%s$BLUE$BOLD]" % self.displayname

    def addHandler(
        self,
        handler: logging.Handler,
        include_descendants: bool = True,
        set_log_level: bool = True,
    ):
        if set_log_level:
            handler.setLevel(self.log_level)
        self._logger.addHandler(handler)
        self._handlers.append(handler)
        if include_descendants:
            self._descendant_handlers.append(handler)
            for child in self.children:
                if isinstance(child, EthenoLogger):
                    child.addHandler(
                        handler,
                        include_descendants=include_descendants,
                        set_log_level=set_log_level,
                    )
                else:
                    child.addHandler(handler)

    def make_logged_file(
        self, prefix=None, suffix=None, mode="w+b", dir: Optional[str] = None
    ):
        """Returns an opened file stream to a unique file created according to the provided naming scheme"""
        if dir is None:
            dir = ""
        else:
            dir = os.path.relpath(
                os.path.realpath(dir), start=os.path.realpath(self.directory)
            )
        os.makedirs(os.path.join(self.directory, dir), exist_ok=True)
        i = 1
        while True:
            if i == 1:
                filename = f"{prefix}{suffix}"
            else:
                filename = f"{prefix}{i}{suffix}"
            path = os.path.join(self.directory, dir, filename)
            if not os.path.exists(path):
                return open(path, mode)
            i += 1

    def make_constant_logged_file(self, contents: Union[str, bytes], *args, **kwargs):
        """Creates a logged file, populates it with the provided contents, and returns the absolute path to the file."""
        if isinstance(contents, str):
            contents = contents.encode("utf-8")
        with self.make_logged_file(*args, **kwargs) as f:
            f.write(contents)  # type: ignore
            return os.path.realpath(f.name)

    def to_log_path(self, absolute_path):
        if self.directory is None:
            return absolute_path
        absolute_path = os.path.realpath(absolute_path)
        dirpath = os.path.realpath(self.directory)
        return os.path.relpath(absolute_path, start=dirpath)

    def save_to_file(self, path, include_descendants=True, log_level=None):
        if log_level is None:
            log_level = self.log_level
        handler = logging.FileHandler(path)
        handler.setLevel(log_level)
        handler.setFormatter(
            logging.Formatter(
                ColorFormatter.remove_color(
                    self.DEFAULT_FORMAT.replace("$NAME", self._name_format())
                ),
                datefmt="%m-%d|%H:%M:%S",
            )
        )
        self.addHandler(
            handler, include_descendants=include_descendants, set_log_level=False
        )

    def save_to_directory(self, path):
        if self.directory == path:
            # we are already set to save to this directory
            return
        elif self.directory is not None:
            raise ValueError(
                "Logger %s's save directory is already set to %s" % (self.name, path)
            )
        self._directory = os.path.realpath(path)
        os.makedirs(path, exist_ok=True)
        self.save_to_file(
            os.path.join(path, "%s.log" % self.name.replace(os.sep, "-")),
            include_descendants=False,
            log_level=DEBUG,
        )
        for child in self.children:
            child.save_to_directory(os.path.join(path, child.name))

    @property
    def log_level(self):
        if self._log_level is None:
            if self.parent is None:
                raise ValueError(
                    "A logger must be provided a parent if `log_level` is None"
                )
            return self.parent.log_level
        else:
            return self._log_level

    @log_level.setter
    def log_level(self, level):
        if not isinstance(level, int):
            try:
                level = getattr(logging, str(level).upper())
            except AttributeError:
                raise ValueError("Invalid log level: %s" % level)
        elif level not in (CRITICAL, ERROR, WARNING, INFO, DEBUG):
            raise ValueError("Invalid log level: %d" % level)
        self._log_level = level
        self._logger.setLevel(level)
        for handler in self._handlers:
            handler.setLevel(level)

    def __getattr__(self, name):
        return getattr(self._logger, name)

    def __repr__(self):
        return f"{type(self).__name__}(name={self.name!r}, log_level={self.log_level!r}, parent={self.parent!r}, cleanup_empty={self.cleanup_empty!r}, displayname={self.displayname!r})"


class StreamLogger(threading.Thread):
    def __init__(self, logger: logging.Logger, *streams, newline_char=b"\n"):
        super().__init__(daemon=True)
        self.logger: logging.Logger = logger
        self.streams = streams
        if isinstance(newline_char, str):
            newline_char = newline_char.encode("utf-8")
        self._newline_char = newline_char
        self._buffers = [b"" for i in range(len(streams))]
        self._done: bool = False
        # TODO: Made a small change here due to the ellipses not being allowed, make sure it does not create any other issues
        self.log: Callable[
            [logging.Logger, Union[str, bytes]], Any
        ] = lambda lgr, message: lgr.info(message)

    def is_done(self) -> bool:
        return self._done

    def run(self):
        while not self.is_done():
            while True:
                got_byte = False
                try:
                    for i, stream in enumerate(self.streams):
                        byte = stream.read(1)
                        while byte is not None and len(byte):
                            if isinstance(byte, str):
                                byte = byte.encode("utf-8")
                            if byte == self._newline_char:
                                self.log(self.logger, self._buffers[i].decode())
                                self._buffers[i] = b""
                            else:
                                self._buffers[i] += byte
                            got_byte = True
                            byte = stream.read(1)
                except Exception:
                    self._done = True
                if not got_byte or self._done:
                    break
            time.sleep(0.5)


class ProcessLogger(StreamLogger):
    def __init__(self, logger, process):
        self.process = process
        super().__init__(
            logger,
            open(process.stdout.fileno(), buffering=1),
            open(process.stderr.fileno(), buffering=1),
        )

    def is_done(self):
        return self.process.poll() is not None


class PtyLogger(StreamLogger):
    def __init__(self, logger, args, cwd=None, **kwargs):
        self.process = ptyprocess.PtyProcessUnicode.spawn(args, cwd=cwd)
        super().__init__(logger, self.process, **kwargs)

    def is_done(self):
        return not self.process.isalive()

    def __getattr__(self, name):
        return getattr(self.process, name)


if __name__ == "__main__":
    logger = EthenoLogger("Testing", DEBUG)
    logger.info("Info")
    logger.critical("Critical")



import pkg_resources
import os
from threading import Thread
from typing import Any, Dict, List, Optional
from werkzeug.serving import make_server

from flask import Flask, jsonify, request, abort
from flask.views import MethodView

from . import logger
from . import threadwrapper
from .client import EthenoClient, JSONRPCError, SelfPostingClient
from .utils import format_hex_address

VERSION: str = pkg_resources.require("etheno")[0].version
VERSION_NAME = f"ToB/v{VERSION}/source/Etheno"
JSONRPC_VERSION = "2.0"
VERSION_ID = 67

app = Flask(__name__)

GETH_DEFAULT_RPC_PORT = 8545
ETH_DEFAULT_RPC_PORT = 8545
PARITY_DEFAULT_RPC_PORT = 8545
PYETHAPP_DEFAULT_RPC_PORT = 4000


def to_account_address(raw_address: int) -> str:
    addr = "%x" % raw_address
    return "0x%s%s" % ("0" * (40 - len(addr)), addr)


_CONTROLLER = threadwrapper.MainThreadController()


class DropPost(RuntimeError):
    pass


OptionalLogger = Optional[logger.EthenoLogger]


class EthenoPlugin:
    _etheno: Optional["Etheno"] = None
    logger: OptionalLogger = None

    @property
    def etheno(self) -> "Etheno":
        return self._etheno

    @etheno.setter
    def etheno(self, instance: "Etheno"):
        if self._etheno is not None:
            if instance is None:
                self._etheno = None
                return
            raise ValueError(
                "An Etheno plugin can only ever be associated with a single Etheno instance"
            )
        self._etheno = instance
        self.logger = logger.EthenoLogger(
            self.__class__.__name__, parent=self._etheno.logger
        )

    @property
    def log_directory(self):
        """Returns a log directory that this client can use to save additional files, or None if one is not available"""
        if self.logger is None:
            return None
        else:
            return self.logger.directory

    def added(self):
        """
        A callback when this plugin is added to an Etheno instance
        """
        pass

    def before_post(self, post_data: Dict[str, Any]) -> Optional[Dict[str, Any]]:
        """
        A callback when Etheno receives a JSON RPC POST, but before it is processed.
        :param post_data: The raw JSON RPC data
        :return: the post_data to be used by Etheno (can be modified); if None, the post proceeds as usual and is not
        modified; if you want to drop the post, `raise DropPost`
        """
        return None

    def after_post(self, post_data: Dict[str, Any], client_results):
        """
        A callback when Etheno receives a JSON RPC POST after it is processed by all clients.
        :param post_data: The raw JSON RPC data
        :param client_results: A lost of the results returned by each client
        """
        pass

    def run(self):
        """
        A callback when Etheno is running and all other clients and plugins are initialized
        """
        pass

    def finalize(self):
        """
        Called when an analysis pass should be finalized (e.g., after a Truffle migration completes).
        Subclasses implementing this function should support it to be called multiple times in a row.
        """
        pass

    def shutdown(self):
        """
        Called before Etheno shuts down.
        The default implementation calls `finalize()`.
        """
        self.finalize()


class Etheno:
    def __init__(self, master_client: Optional[SelfPostingClient] = None):
        self.accounts = []
        self._master_client: Optional[SelfPostingClient] = None
        if master_client is None:
            self.master_client = None
        else:
            self.master_client = master_client
        self.clients: List[EthenoClient] = []
        self.rpc_client_result = None
        self.plugins: List[EthenoPlugin] = []
        self._shutting_down: bool = False
        self.logger: logger.EthenoLogger = logger.EthenoLogger("Etheno", logger.INFO)

    @property
    def log_level(self) -> int:
        return self.logger.log_level

    @log_level.setter
    def log_level(self, level: int):
        self.logger.log_level = level

    @property
    def master_client(self) -> Optional[SelfPostingClient]:
        return self._master_client

    @master_client.setter
    def master_client(self, client: Optional[SelfPostingClient]):
        if client is None:
            if self._master_client is not None:
                self._master_client.etheno = None
                self._master_client = None
            return
        if not isinstance(client, SelfPostingClient):
            raise Exception(
                "The master client must be an instance of a SelfPostingClient"
            )
        client.etheno = self
        self._master_client = client
        self.accounts: list[int] = list(
            map(
                lambda a: int(a, 16),
                client.post({"id": 1, "jsonrpc": "2.0", "method": "eth_accounts"})[
                    "result"
                ],
            )
        )
        for client in self.clients:
            self._create_accounts(client)

    def estimate_gas(self, transaction) -> Optional[int]:
        """
        Estimates the gas cost of a transaction.
        Iterates through all clients until it finds a client that is capable of estimating the gas cost without error.
        If all clients return an error, this function will return None.
        """
        clients = [self.master_client] + [
            client for client in self.clients if hasattr(client, "estimate_gas")
        ]  # type: ignore
        for client in clients:
            try:
                return client.estimate_gas(transaction)
            except JSONRPCError:
                continue
        return None

    def post(self, data):
        self.logger.debug(f"Handling JSON RPC request {data}")

        for plugin in self.plugins:
            try:
                new_data = plugin.before_post(dict(data))
                if new_data is not None and new_data != data:
                    self.logger.debug(
                        f"Incoming JSON RPC request {data} changed by plugin {plugin!r} to {new_data}"
                    )
                    data = new_data
            except DropPost:
                self.logger.info(
                    f"Incoming JSON RPC request {data} dropped by plugin {plugin!r}"
                )

        method = data["method"]
        args = ()
        kwargs = {}
        if "params" in data:
            params = data["params"]
            if len(params) == 1 and isinstance(params[0], dict):
                kwargs = dict(params[0])
                # handle Python reserved words:
                if "from" in kwargs:
                    kwargs["from_addr"] = kwargs["from"]
                    del kwargs["from"]
            else:
                args = data["params"]

        if self.master_client is None:
            ret = None
        else:
            if method == "eth_getTransactionReceipt":
                # for eth_getTransactionReceipt, make sure we block until all clients have mined the transaction
                ret = self.master_client.wait_for_transaction(data["params"][0])
                if "id" in data and "id" in ret:
                    ret["id"] = data["id"]
            else:
                try:
                    ret = self.master_client.post(data)
                except JSONRPCError as e:
                    self.logger.error(e)
                    ret = e

        self.rpc_client_result = ret
        self.logger.debug(
            f"Result from the master client ({self.master_client}): {ret}"
        )

        results = []

        for client in self.clients:
            try:
                if hasattr(client, method):
                    self.logger.info(
                        "Enrobing JSON RPC call to %s.%s" % (client, method)
                    )
                    function = getattr(client, method)
                    if function is not None:
                        kwargs["rpc_client_result"] = ret
                        results.append(function(*args, **kwargs))
                    else:
                        self.logger.warn(f"Function {method} of {client} is None!")
                        results.append(None)
                elif isinstance(client, SelfPostingClient):
                    if method == "eth_getTransactionReceipt":
                        # for eth_getTransactionReceipt, make sure we block until all clients have mined the transaction
                        results.append(client.wait_for_transaction(data["params"][0]))
                    else:
                        results.append(client.post(data))
                else:
                    results.append(None)
            except JSONRPCError as e:
                self.logger.error(e)
                results.append(e)
            self.logger.debug(f"Result from client {client}: {results[-1]}")

        if ret is None:
            return None

        results = [ret] + results
        for plugin in self.plugins:
            plugin.after_post(data, results)

        return ret

    def add_plugin(self, plugin: EthenoPlugin):
        plugin.etheno = self
        self.plugins.append(plugin)
        plugin.added()

    def remove_plugin(self, plugin: EthenoPlugin):
        """Removes a plugin, automatically calling plugin.shutdown() in the process

        :param plugin: The plugin to remove
        """
        self.plugins.remove(plugin)
        plugin.shutdown()
        plugin.etheno = None

    def _create_accounts(self, client):
        for account in self.accounts:
            # TODO: Actually get the correct balance from the JSON RPC client instead of using hard-coded 100.0 ETH
            client.create_account(balance=int(100.0 * 10**18), address=account)

    def add_client(self, client: EthenoClient):
        client.etheno = self
        self.clients.append(client)
        self._create_accounts(client)

    def deploy_contract(
        self, from_address, bytecode, gas=0x99999, gas_price=None, value=0
    ) -> Optional[int]:
        if gas_price is None:
            gas_price = self.master_client.get_gas_price()
        if isinstance(bytecode, bytes):
            bytecode = bytecode.decode()
        if not bytecode.startswith("0x"):
            bytecode = "0x%s" % bytecode
        tx_hash = self.post(
            {
                "id": 1,
                "jsonrpc": "2.0",
                "method": "eth_sendTransaction",
                "params": [
                    {
                        "from": format_hex_address(from_address, True),
                        "gas": "0x%x" % gas,
                        "gasPrice": "0x%x" % gas_price,
                        "value": "0x0",
                        "data": bytecode,
                    }
                ],
            }
        )["result"]
        receipt = self.master_client.wait_for_transaction(tx_hash)
        if (
            "result" in receipt
            and receipt["result"]
            and "contractAddress" in receipt["result"]
            and receipt["result"]["contractAddress"]
        ):
            return int(receipt["result"]["contractAddress"], 16)
        else:
            return None

    def get_transaction_receipt_request(self, tx_hash: str) -> Dict:
        """
        Takes in a transaction hash and returns the request object for an eth_getTransactionReceipt API call
        """
        return {
            "id": 1,
            "jsonrpc": "2.0",
            "method": "eth_getTransactionReceipt",
            "params": [tx_hash],
        }

    def shutdown(self, port: int = GETH_DEFAULT_RPC_PORT):
        if self._shutting_down:
            return
        self._shutting_down = True
        for plugin in self.plugins:
            plugin.shutdown()
        # Send a web request to the server to shut down:
        if self.master_client:
            self.master_client.shutdown()
        for client in self.clients:
            client.shutdown()
        self.logger.close()
        _CONTROLLER.quit()

    def run(self, debug=True, run_publicly=False, port=GETH_DEFAULT_RPC_PORT):
        # Manticore only works in the main thread, so use a threadsafe wrapper:
        def server_thread():
            IS_DOCKER = os.environ.get("DOCKER", 0)
            if run_publicly or IS_DOCKER:
                host = "0.0.0.0"
            else:
                host = "127.0.0.1"
            # Do not use the reloader, because Flask needs to run in the main thread to use the reloader
            server = make_server(host=host, port=port, app=app, threaded=True)
            return server
            # app.run(debug=debug, host=host, port=port, use_reloader=False)

        server = server_thread()
        thread = Thread(target=server.serve_forever)
        thread.start()
        self.logger.info("Etheno v%s" % VERSION)

        for plugin in self.plugins:
            plugin.run()

        _CONTROLLER.run()
        self.shutdown()
        self.logger.info("Shutting Etheno down")
        server.shutdown()
        thread.join()


ETHENO = Etheno()


class EthenoView(MethodView):
    def post(self):
        data = request.get_json()
        was_list = False

        if isinstance(data, list):
            if len(data) == 1:
                was_list = True
                data = data[0]
            else:
                ETHENO.logger.error("Unexpected POST data: %s" % data)
                abort(400)

        if "jsonrpc" not in data or "method" not in data:
            abort(400)
        try:
            jsonrpc_version = float(data["jsonrpc"])
        except ValueError:
            abort(400)
        if jsonrpc_version < 2.0:
            abort(426)
        elif jsonrpc_version > 2.0:
            ETHENO.logger.warn(
                f"Client is using a newer version of the JSONRPC protocol! Expected 2.0, but got {jsonrpc_version}"
            )

        ret = ETHENO.post(data)

        ETHENO.logger.debug(f"Returning {ret}")

        if ret is None:
            return None

        if isinstance(ret, JSONRPCError):
            ret = ret.result

        if was_list:
            ret = [ret]
        ret = jsonify(ret)

        return ret



from collections.abc import Sequence
import shlex
import time

from typing import Iterable

from .logger import EthenoLogger, PtyLogger


def make_list(args: Iterable):
    if isinstance(args, str):
        return shlex.split(args)
    elif isinstance(args, Sequence) and not isinstance(args, bytes):
        if isinstance(args, list):
            return args
        else:
            return list(args)
    else:
        return [args]


class Truffle(object):
    def __init__(self, truffle_cmd="truffle", parent_logger=None, log_level=None):
        self._running = False
        self.logger = EthenoLogger("Truffle", log_level=log_level, parent=parent_logger)
        self.truffle_cmd = make_list(truffle_cmd)

    def terminate(self):
        self._running = False

    def run_tests(self):
        return self.run("test")

    def run_migrate(self):
        return self.run("migrate")

    def run(self, args):
        self._running = True
        args = make_list(args)

        p = PtyLogger(self.logger, ["/usr/bin/env"] + self.truffle_cmd + args)
        p.start()

        try:
            while p.isalive():
                if not self._running:
                    self.logger.info(
                        "Etheno received a shutdown signal; terminating truffle %s"
                        % " ".join(args)
                    )
                    break
                time.sleep(1.0)
        except KeyboardInterrupt as e:
            self.logger.info(
                "Caught keyboard interrupt; terminating truffle %s" % " ".join(args)
            )
            raise e
        finally:
            p.close(force=True)
        return p.wait()



import argparse
import json
import os
import shlex
import sys
from threading import Thread

from .client import RpcProxyClient
from .differentials import DifferentialTester
from .etheno import app, EthenoView, GETH_DEFAULT_RPC_PORT, ETHENO, VERSION_NAME
from .genesis import Account, make_accounts, make_genesis
from .jsonrpc import EventSummaryExportPlugin, JSONRPCExportPlugin
from .synchronization import AddressSynchronizingClient, RawTransactionClient
from .utils import (
    clear_directory,
    decode_value,
    find_open_port,
    format_hex_address,
    ynprompt,
)
from . import ganache
from . import geth
from . import logger
from . import parity
from . import truffle

# Constant for converting whole units to wei
ETHER = 1e18


def main(argv=None):
    parser = argparse.ArgumentParser(
        description="An Ethereum JSON RPC multiplexer, differential fuzzer, and test framework integration tool."
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        default=False,
        help="Enable debugging from within the web server",
    )
    parser.add_argument(
        "--run-publicly",
        action="store_true",
        default=False,
        help="Allow the web server to accept external connections",
    )
    parser.add_argument(
        "-p",
        "--port",
        type=int,
        default=GETH_DEFAULT_RPC_PORT,
        help="Port on which to run the JSON RPC webserver (default=%d)"
        % GETH_DEFAULT_RPC_PORT,
    )
    parser.add_argument(
        "-a",
        "--accounts",
        type=int,
        default=10,
        help="Number of accounts to create in the client (default=10)",
    )
    parser.add_argument(
        "-b",
        "--balance",
        type=float,
        default=1000.0,
        help="Default balance (in Ether) to seed to each account (default=100.0)",
    )
    # TODO: do we really need a gas price specified for ganache? is there a use case here?
    parser.add_argument(
        "-c",
        "--gas-price",
        type=int,
        default=20000000000,
        help="Default gas price (default=20000000000)",
    )
    # TODO: networkID can have a default value it seems like
    parser.add_argument(
        "-i",
        "--network-id",
        type=int,
        default=None,
        help="Specify a network ID (default is the network ID of the master client)",
    )
    parser.add_argument(
        "-t",
        "--truffle",
        action="store_true",
        default=False,
        help="Run the truffle migrations in the current directory and exit",
    )
    parser.add_argument(
        "--truffle-cmd",
        type=str,
        default="truffle",
        help="Command to run truffle (default=truffle)",
    )
    parser.add_argument(
        "--truffle-args",
        type=str,
        default="migrate",
        help="Arguments to pass to truffle (default=migrate)",
    )
    parser.add_argument(
        "-g",
        "--ganache",
        action="store_true",
        default=False,
        help="Run Ganache as a master JSON RPC client (cannot be used in conjunction with --master)",
    )
    # TODO: This cmd-line argument is error-prone and should probably be removed. Commenting it out for now
    """
    parser.add_argument('--ganache-cmd', type=str, default=None, help='Specify a command that runs Ganache '
                                                                      '(default="/usr/bin/env ganache")')
    """
    parser.add_argument(
        "--ganache-args",
        type=str,
        default=None,
        help="Additional arguments to pass to Ganache",
    )
    parser.add_argument(
        "--ganache-port",
        type=int,
        default=None,
        help="Port on which to run Ganache (defaults to the closest available port to the port "
        "specified with --port plus one)",
    )
    parser.add_argument(
        "-go",
        "--geth",
        action="store_true",
        default=False,
        help="Run Geth as a JSON RPC client",
    )
    parser.add_argument(
        "--geth-port",
        type=int,
        default=None,
        help="Port on which to run Geth (defaults to the closest available port to the port specified "
        "with --port plus one)",
    )
    parser.add_argument(
        "-pa",
        "--parity",
        action="store_true",
        default=False,
        help="Run Parity as a JSON RPC client",
    )
    parser.add_argument(
        "--parity-port",
        type=int,
        default=None,
        help="Port on which to run Parity (defaults to the closest available port to the port "
        "specified with --port plus one)",
    )
    parser.add_argument(
        "-j",
        "--genesis",
        type=str,
        default=None,
        help="Path to a genesis.json file to use for initializing clients. Any genesis-related options "
        "like --network-id will override the values in this file. If --accounts is greater than "
        "zero, that many new accounts will be appended to the accounts in the genesis file.",
    )
    parser.add_argument(
        "--save-genesis",
        type=str,
        default=None,
        help="Save a genesis.json file to reproduce the state of this run. Note that this genesis file "
        "will include all known private keys for the genesis accounts, so use this with caution.",
    )
    parser.add_argument(
        "--constantinople-block",
        type=int,
        default=None,
        help="The block in which to enable Constantinople EIPs (default=do not enable Constantinople)",
    )
    parser.add_argument(
        "--constantinople",
        action="store_true",
        default=False,
        help="Enables Constantinople EIPs; equivalent to `--constantinople-block 0`",
    )
    parser.add_argument(
        "--no-differential-testing",
        action="store_false",
        dest="run_differential",
        default=True,
        help="Do not run differential testing, which is run by default",
    )
    parser.add_argument(
        "-l",
        "--log-level",
        type=str.upper,
        choices={"CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"},
        default="INFO",
        help="Set Etheno's log level (default=INFO)",
    )
    parser.add_argument(
        "--log-file",
        type=str,
        default=None,
        help="Path to save all log output to a single file",
    )
    parser.add_argument(
        "--log-dir",
        type=str,
        default=None,
        help="Path to a directory in which to save all log output, divided by logging source",
    )
    parser.add_argument(
        "-d",
        "--dump-jsonrpc",
        type=str,
        default=None,
        help="Path to a JSON file in which to dump all raw JSON RPC calls; if `--log-dir` is provided, "
        "the raw JSON RPC calls will additionally be dumped to `rpc.json` in the log directory.",
    )
    parser.add_argument(
        "-x",
        "--export-summary",
        type=str,
        default=None,
        help="Path to a JSON file in which to export an event summary",
    )
    parser.add_argument(
        "-v",
        "--version",
        action="store_true",
        default=False,
        help="Print version information and exit",
    )
    parser.add_argument(
        "client",
        type=str,
        nargs="*",
        help="JSON RPC client URLs to multiplex; if no client is specified for --master, the first "
        'client in this list will default to the master (format="http://foo.com:8545/")',
    )
    parser.add_argument(
        "-s",
        "--master",
        type=str,
        default=None,
        help="A JSON RPC client to use as the master "
        '(format="http://foo.com:8545/")',
    )
    parser.add_argument(
        "--raw",
        type=str,
        nargs="*",
        action="append",
        help="JSON RPC client URLs to multiplex that do not have any local accounts; Etheno will "
        "automatically use auto-generated accounts with known private keys, pre-sign all "
        "transactions, and only use eth_sendRawTransaction",
    )

    if argv is None:
        argv = sys.argv

    args = parser.parse_args(argv[1:])

    if args.version:
        print(VERSION_NAME)
        sys.exit(0)

    if args.constantinople and args.constantinople_block is None:
        args.constantinople_block = 0

    ETHENO.log_level = args.log_level

    if args.log_file:
        ETHENO.logger.save_to_file(args.log_file)

    if args.log_dir:
        if os.path.exists(args.log_dir):
            if not ynprompt(
                "Logging path `%s` already exists! Would you like to overwrite it? [yN] "
                % args.log_dir
            ):
                sys.exit(1)
            elif os.path.isfile(args.log_dir):
                os.remove(args.log_dir)
            else:
                # don't delete the directory, just its contents
                # we can't use shutil.rmtree here, because that deletes the directory and also it doesn't work on
                # symlinks
                if not ynprompt(
                    "We are about to delete the contents of `%s`. Are you sure? [yN] "
                    % args.log_dir
                ):
                    sys.exit(1)
                abspath = os.path.abspath(args.log_dir)
                if (
                    abspath == ""
                    or abspath == "/"
                    or abspath.endswith("://")
                    or abspath.endswith(":\\\\")
                ):
                    print(
                        "Wait a sec, you want me to delete `%s`?!\nThat looks too dangerous.\nIf I were to do that, "
                        "you'd file an angry GitHub issue complaining that I deleted your hard drive.\nYou're on "
                        "your own deleting this directory!" % abspath
                    )
                    sys.exit(1)
                clear_directory(args.log_dir)

        ETHENO.logger.save_to_directory(args.log_dir)
        if not args.log_file:
            # Also create a unified log in the log dir:
            ETHENO.logger.save_to_file(os.path.join(args.log_dir, "Complete.log"))

        ETHENO.add_plugin(JSONRPCExportPlugin(os.path.join(args.log_dir, "rpc.json")))

    if args.dump_jsonrpc is not None:
        ETHENO.add_plugin(JSONRPCExportPlugin(args.dump_jsonrpc))

    if args.export_summary is not None:
        ETHENO.add_plugin(EventSummaryExportPlugin(args.export_summary))

    if args.genesis is None:
        # Set defaults since no genesis was supplied
        if args.accounts is None:
            args.accounts = 10
        if args.gas_price is None:
            args.gas_price = 20000000000

    accounts = []

    # TODO: args.gas_price is not set if a genesis file is provided
    if args.genesis:
        with open(args.genesis, "rb") as f:
            genesis = json.load(f)
            if "config" not in genesis:
                genesis["config"] = {}
            if "alloc" not in genesis:
                genesis["alloc"] = {}
            if args.network_id is None:
                args.network_id = genesis["config"].get("chainId", None)
            if args.constantinople_block is None:
                args.constantinople_block = genesis["config"].get(
                    "constantinopleBlock", None
                )
                args.constantinople = args.constantinople_block is not None
            for addr, bal in genesis["alloc"].items():
                pkey = None
                if "privateKey" in bal:
                    pkey = bal["privateKey"]
                accounts.append(
                    Account(
                        address=int(addr, 16),
                        balance=decode_value(bal["balance"]),
                        private_key=decode_value(pkey),
                    )
                )
    else:
        # We will generate it further below once we've resolved all of the parameters
        genesis = None

    accounts += make_accounts(args.accounts, default_balance=int(args.balance * ETHER))

    if genesis is not None:
        # add the new accounts to the genesis
        for account in accounts[len(genesis["alloc"]) :]:
            genesis["alloc"][format_hex_address(account.address)] = {
                "balance": "%d" % account.balance,
                "privateKey": format_hex_address(account.private_key),
                "comment": "`privateKey` and `comment` are ignored.  In a real chain, the private key should _not_ be "
                "stored!",
            }

    if args.raw is None:
        args.raw = []
    else:
        args.raw = [r[0] for r in args.raw]

    # TODO: This if/elif/else logic is flawed - needs rework
    if args.ganache and args.master:
        parser.print_help()
        sys.stderr.write(
            "\nError: You cannot specify both --ganache and --master at the same time!\n"
        )
        sys.exit(1)
    elif args.ganache:
        if args.ganache_port is None:
            args.ganache_port = find_open_port(args.port + 1)

        if args.network_id is None:
            args.network_id = 0x657468656E6F  # 'etheno' in hex

        # Have to use hex() so that string is hex-encoded (prefixed with 0x) that is necessary for Ganache v7.0+
        # https://github.com/trufflesuite/ganache/discussions/1075
        ganache_accounts = [
            "--account=%s,0x%x" % (hex(acct.private_key), acct.balance)
            for acct in accounts
        ]

        ganache_args = ganache_accounts + [
            "-g",
            str(args.gas_price),
            "-i",
            str(args.network_id),
        ]

        if args.ganache_args is not None:
            ganache_args += shlex.split(args.ganache_args)
        # Removed cmd argument
        ganache_instance = ganache.Ganache(args=ganache_args, port=args.ganache_port)

        ETHENO.master_client = ganache.GanacheClient(ganache_instance)

        ganache_instance.start()
    elif args.master:
        ETHENO.master_client = AddressSynchronizingClient(RpcProxyClient(args.master))
    elif args.client and not args.geth and not args.parity:
        ETHENO.master_client = AddressSynchronizingClient(
            RpcProxyClient(args.client[0])
        )
        args.client = args.client[1:]
    elif args.raw and not args.geth and not args.parity:
        ETHENO.master_client = RawTransactionClient(
            RpcProxyClient(args.raw[0]), accounts
        )
        args.raw = args.raw[1:]

    if args.network_id is None:
        if ETHENO.master_client:
            args.network_id = int(
                ETHENO.master_client.post(
                    {"id": 1, "jsonrpc": "2.0", "method": "net_version"}
                )["result"],
                16,
            )
        else:
            args.network_id = 0x657468656E6F  # 'etheno' in hex

    if genesis is None:
        genesis = make_genesis(
            network_id=args.network_id,
            accounts=accounts,
            constantinople_block=args.constantinople_block,
        )
    else:
        # Update the genesis with any overridden values
        genesis["config"]["chainId"] = args.network_id

    if args.save_genesis:
        with open(args.save_genesis, "wb") as f:
            f.write(json.dumps(genesis).encode("utf-8"))
            ETHENO.logger.info("Saved genesis to %s" % args.save_genesis)

    if args.geth:
        if args.geth_port is None:
            args.geth_port = find_open_port(args.port + 1)

        geth_instance = geth.GethClient(genesis=genesis, port=args.geth_port)
        geth_instance.etheno = ETHENO
        for account in accounts:
            # TODO: Make some sort of progress bar here
            geth_instance.logger.info(
                "Unlocking Geth account %s" % format_hex_address(account.address, True)
            )
            geth_instance.import_account(account.private_key)
        geth_instance.start(unlock_accounts=True)
        if ETHENO.master_client is None:
            ETHENO.master_client = geth_instance
        else:
            ETHENO.add_client(AddressSynchronizingClient(geth_instance))

    if args.parity:
        if args.parity_port is None:
            if args.geth_port is not None:
                args.parity_port = find_open_port(args.geth_port + 1)
            else:
                args.parity_port = find_open_port(args.port + 1)

        parity_instance = parity.ParityClient(genesis=genesis, port=args.parity_port)
        parity_instance.etheno = ETHENO
        for account in accounts:
            # TODO: Make some sort of progress bar here
            parity_instance.import_account(account.private_key)
        parity_instance.start(unlock_accounts=True)
        if ETHENO.master_client is None:
            ETHENO.master_client = parity_instance
        else:
            ETHENO.add_client(AddressSynchronizingClient(parity_instance))

    for client in args.client:
        ETHENO.add_client(AddressSynchronizingClient(RpcProxyClient(client)))

    for client in args.raw:
        ETHENO.add_client(RawTransactionClient(RpcProxyClient(client), accounts))

    if args.truffle:
        truffle_controller = truffle.Truffle(
            truffle_cmd=args.truffle_cmd, parent_logger=ETHENO.logger
        )

        def truffle_thread():
            if ETHENO.master_client:
                ETHENO.master_client.wait_until_running()
            ETHENO.logger.info("Etheno Started! Running Truffle...")
            ret = truffle_controller.run(args.truffle_args)
            if ret != 0:
                ETHENO.logger.error("Truffle exited with code %s" % ret)
                ETHENO.shutdown()
                # TODO: Propagate the error code elsewhere so Etheno doesn't exit with code 0

            for plugin in ETHENO.plugins:
                plugin.finalize()

            if not ETHENO.clients and not ETHENO.plugins:
                ETHENO.logger.info("No clients or plugins running; exiting...")
                ETHENO.shutdown()

        thread = Thread(target=truffle_thread)
        thread.start()

    # Without Manticore integration the only client types are geth, parity, and command-line raw/regular clients.
    # So checking len() >= 1 should be sufficient.
    if (
        args.run_differential
        and (ETHENO.master_client is not None)
        and len(ETHENO.clients) >= 1
    ):
        # There are at least two non-Manticore clients running
        ETHENO.logger.info(
            "Initializing differential tests to compare clients %s"
            % ", ".join(map(str, [ETHENO.master_client] + ETHENO.clients))
        )
        ETHENO.add_plugin(DifferentialTester())

    had_plugins = len(ETHENO.plugins) > 0

    if ETHENO.master_client is None and not ETHENO.clients and not ETHENO.plugins:
        if not had_plugins:
            ETHENO.logger.info("No clients or plugins provided; exiting...")
        # else: this can also happen if there were plugins but they uninstalled themselves after running
        return

    etheno = EthenoView()
    app.add_url_rule("/", view_func=etheno.as_view("etheno"))

    ETHENO.run(debug=args.debug, run_publicly=args.run_publicly, port=args.port)
    if args.truffle:
        truffle_controller.terminate()

    if args.log_file is not None:
        print("Log file saved to: %s" % os.path.realpath(args.log_file))
    if args.log_dir is not None:
        print(
            "Logs %ssaved to: %s"
            % (["", "also "][args.log_file is not None], os.path.realpath(args.log_dir))
        )


if __name__ == "__main__":
    main()



from .client import RpcProxyClient
from .etheno import EthenoPlugin
from .utils import format_hex_address

# TODO: what is this file for?
class ContractSynchronizer(EthenoPlugin):
    def __init__(self, source_client, contract_address):
        if isinstance(source_client, str):
            source_client = RpcProxyClient(source_client)
        self.source = source_client
        self.contract = format_hex_address(contract_address, True)

    def added(self):
        # get the contract:
        pass



import json
from typing import Dict, TextIO, Union

from .etheno import EthenoPlugin
from .utils import format_hex_address
from .client import JSONRPCError

# source: https://ethereum.stackexchange.com/a/83855
import rlp
from eth_typing import HexStr
from eth_utils import keccak, to_bytes
from rlp.sedes import Binary, big_endian_int, binary
from web3 import Web3
from web3.auto import w3


class Transaction(rlp.Serializable):
    fields = [
        ("nonce", big_endian_int),
        ("gas_price", big_endian_int),
        ("gas", big_endian_int),
        ("to", Binary.fixed_length(20, allow_empty=True)),
        ("value", big_endian_int),
        ("data", binary),
        ("v", big_endian_int),
        ("r", big_endian_int),
        ("s", big_endian_int),
    ]


def hex_to_bytes(data: str) -> bytes:
    return to_bytes(hexstr=HexStr(data))


def decode_raw_tx(raw_tx: str):
    tx_bytes = hex_to_bytes(raw_tx)
    tx = rlp.decode(tx_bytes, Transaction)
    hash_tx = Web3.toHex(keccak(tx_bytes))
    from_ = w3.eth.account.recover_transaction(raw_tx)
    to = w3.toChecksumAddress(tx.to) if tx.to else None
    data = w3.toHex(tx.data)
    r = hex(tx.r)
    s = hex(tx.s)
    chain_id = (tx.v - 35) // 2 if tx.v % 2 else (tx.v - 36) // 2
    return {
        "txHash": hash_tx,
        "from": from_,
        "to": to,
        "nonce": tx.nonce,
        "gas": hex(tx.gas),
        "gasPrice": hex(tx.gas_price),
        "value": hex(tx.value),
        "data": data,
        "chainId": chain_id,
        "r": r,
        "s": s,
        "v": tx.v,
    }


class JSONExporter:
    def __init__(self, out_stream: Union[str, TextIO]):
        self._was_path = isinstance(out_stream, str)
        if self._was_path:
            self.output = open(out_stream, "w", encoding="utf8")
        else:
            self.output = out_stream
        self.output.write("[")
        self._count = 0
        self._finalized = False

    def finalize(self):
        if self._finalized:
            return
        if self._count:
            self.output.write("\n")
        self.output.write("]")
        self.output.flush()
        if self._was_path:
            self.output.close()
        self._finalized = True

    def write_entry(self, entry):
        if self._finalized:
            return
        if self._count > 0:
            self.output.write(",")
        self._count += 1
        self.output.write("\n")
        json.dump(entry, self.output)
        self.output.flush()


class JSONRPCExportPlugin(EthenoPlugin):
    def __init__(self, out_stream: Union[str, TextIO]):
        self._exporter = JSONExporter(out_stream)

    def after_post(self, post_data, client_results):
        self._exporter.write_entry([post_data, client_results])

    def finalize(self):
        self._exporter.finalize()
        if hasattr(self._exporter.output, "name"):
            self.logger.info(
                f"Raw JSON RPC messages dumped to {self._exporter.output.name}"
            )


class EventSummaryPlugin(EthenoPlugin):
    def __init__(self):
        self._transactions: Dict[
            int, Dict[str, object]
        ] = {}  # Maps transaction hashes to their eth_sendTransaction arguments

    def handle_contract_created(
        self,
        creator_address: str,
        contract_address: str,
        gas_used: str,
        gas_price: str,
        data: str,
        value: str,
    ):
        self.logger.info(
            f"Contract created at {contract_address} with {(len(data)-2)//2} bytes of data by account {creator_address} for {gas_used} gas with a gas price of {gas_price}"
        )

    def handle_function_call(
        self,
        from_address: str,
        to_address: str,
        gas_used: str,
        gas_price: str,
        data: str,
        value: str,
    ):
        self.logger.info(
            f"Function call with {value} wei from {from_address} to {to_address} with {(len(data)-2)//2} bytes of data for {gas_used} gas with a gas price of {gas_price}"
        )

    def handle_unlogged_transactions(self):
        unlogged_transactions = dict(
            filter(lambda txn: txn[1]["is_logged"] == False, self._transactions.items())
        )
        for (tx_hash, txn) in unlogged_transactions.items():
            post_data = self._etheno.get_transaction_receipt_request(tx_hash)
            self._etheno.post(post_data)

    def after_post(self, post_data, result):
        if len(result):
            result = result[0]
        if "method" not in post_data:
            return
        # Fixes bug that occurs when a JSONRPCError is attempted to be logged
        if isinstance(result, JSONRPCError):
            self.logger.info(
                f"Received a JSON RPC Error when logging transaction...skipping event logging"
            )
            return

        elif (
            post_data["method"] == "eth_sendTransaction"
            or post_data["method"] == "eth_sendRawTransaction"
        ) and "result" in result:
            transaction_hash = result["result"]
            # Add a boolean to check at shutdown whether everything has been logged.
            if post_data["method"] == "eth_sendRawTransaction":
                self._transactions[transaction_hash] = {
                    "transaction": decode_raw_tx(post_data["params"][0]),
                    "is_logged": False,
                }
            else:
                self._transactions[transaction_hash] = {
                    "transaction": post_data["params"][0],
                    "is_logged": False,
                }
        elif post_data["method"] == "evm_mine":
            self.handle_increase_block_number()
        elif post_data["method"] == "evm_increaseTime":
            self.handle_increase_block_timestamp(post_data["params"][0])
        elif post_data["method"] == "eth_getTransactionReceipt":
            transaction_hash = post_data["params"][0]
            if transaction_hash not in self._transactions:
                self.logger.error(
                    f'Received transaction receipt {result} for unknown transaction hash {post_data["params"][0]}'
                )
                return
            (original_transaction, is_logged) = (
                self._transactions[transaction_hash]["transaction"],
                self._transactions[transaction_hash]["is_logged"],
            )
            # Check if it was logged already
            if is_logged:
                self.logger.debug(
                    f"Transaction hash {transaction_hash} has already been logged. This should not happen."
                )
                return
            if (
                "value" not in original_transaction
                or original_transaction["value"] is None
            ):
                value = "0x0"
            else:
                value = original_transaction["value"]
            if "to" not in result["result"] or result["result"]["to"] is None:
                # this transaction is creating a contract:
                # TODO: key errors are likely here...need to figure out a better way to do error handling
                # TODO: log a warning about non-zero ether values
                contract_address = result["result"]["contractAddress"]
                self.handle_contract_created(
                    original_transaction["from"],
                    contract_address,
                    result["result"]["gasUsed"],
                    result["result"]["effectiveGasPrice"],
                    original_transaction["data"],
                    value,
                )
            else:
                self.handle_function_call(
                    original_transaction["from"],
                    original_transaction["to"],
                    result["result"]["gasUsed"],
                    result["result"]["effectiveGasPrice"],
                    original_transaction["data"]
                    if "data" in original_transaction
                    else "0x",
                    value,
                )
            # Transaction has been logged successfully
            self._transactions[transaction_hash]["is_logged"] = True


class EventSummaryExportPlugin(EventSummaryPlugin):
    def __init__(self, out_stream: Union[str, TextIO]):
        super().__init__()
        self._exporter = JSONExporter(out_stream)

    def run(self):
        for address in self.etheno.accounts:
            self._exporter.write_entry(
                {"event": "AccountCreated", "address": format_hex_address(address)}
            )
        super().run()

    def handle_increase_block_number(self):
        self._exporter.write_entry(
            {"event": "BlockMined", "number_increment": "1", "timestamp_increment": "0"}
        )

    def handle_increase_block_timestamp(self, number: str):
        self._exporter.write_entry(
            {
                "event": "BlockMined",
                "number_increment": "0",
                "timestamp_increment": str(number),
            }
        )

    def handle_contract_created(
        self,
        creator_address: str,
        contract_address: str,
        gas_used: str,
        gas_price: str,
        data: str,
        value: str,
    ):
        self._exporter.write_entry(
            {
                "event": "ContractCreated",
                "from": creator_address,
                "contract_address": contract_address,
                "gas_used": gas_used,
                "gas_price": gas_price,
                "data": data,
                "value": value,
            }
        )
        super().handle_contract_created(
            creator_address, contract_address, gas_used, gas_price, data, value
        )

    def handle_function_call(
        self,
        from_address: str,
        to_address: str,
        gas_used: str,
        gas_price: str,
        data: str,
        value: str,
    ):
        self._exporter.write_entry(
            {
                "event": "FunctionCall",
                "from": from_address,
                "to": to_address,
                "gas_used": gas_used,
                "gas_price": gas_price,
                "data": data,
                "value": value,
            }
        )
        super().handle_function_call(
            from_address, to_address, gas_used, gas_price, data, value
        )

    def finalize(self):
        super().handle_unlogged_transactions()
        self._exporter.finalize()
        if hasattr(self._exporter.output, "name"):
            self.logger.info(
                f"Event summary JSON saved to {self._exporter.output.name}"
            )
